"""GATK utilities"""
from pyppl import Proc
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

#############################
# GATK utilities            #
#############################

_sReferenceFai = """
# make sure reference index is generated
if [[ ! -e "{{brings.reffile#fai}}" ]]; then
	echo "[BioProcs] Cannot find index file for reference: {{reffile.orig}}, generating ..." 1>&2
	({{args.samtools}} faidx "{{reffile | realpath}}" ]] && \\
		ln -s "{{reffile | realpath}}.fai" "{{reffile}}.fai" && \\
		echo "[BioProcs] Generated: {{reffile | realpath}}.fai" 1>&2) || \\
	({{args.samtools}} faidx "{{reffile.orig}}" ]] && \\
		ln -s "{{reffile.orig}}.fai" "{{reffile}}.fai" && \\
		echo "[BioProcs] Generated: {{reffile.orig}}.fai" 1>&2) || \\
	({{args.samtools}} faidx "{{reffile}}" && \\
		echo "[BioProcs] Generated: {{reffile}}.fai" 1>&2)
fi
"""

_sReferenceDict = """
# make sure reference dict is generated
if [[ ! -e "{{brings.reffile#dict}}" ]]; then
	echo "[BioProcs] Cannot find dict file for reference: {{reffile.orig}}, generating ..." 1>&2
	({{args.picard}} CreateSequenceDictionary R="{{reffile}}" O="{{reffile | realpath | prefix}}.dict" && \\
		ln -s "{{reffile | realpath | prefix}}.dict" "{{reffile | prefix}}.dict" && \\
		echo "[BioProcs] Generated: {{reffile | realpath | prefix}}.dict" 1>&2) || \\
	({{args.picard}} CreateSequenceDictionary R="{{reffile}}" O="{{reffile.orig | prefix}}.dict" && \\
		ln -s "{{reffile.orig | prefix}}.dict" "{{reffile | prefix}}.dict" && \\
		echo "[BioProcs] Generated: {{reffile.orig | prefix}}.dict" 1>&2) || \\
	({{args.picard}} CreateSequenceDictionary R="{{reffile}}" O="{{reffile | prefix}}.dict" && \\
		echo "[BioProcs] Generated: {{reffile | prefix}}.dict" 1>&2)
fi
"""

_sBamBai = """
# get index file of bam file
if [[ ! -e "{{brings.bamfile}}" ]]; then
	echo "[BioProcs] Cannot find index file for input file: {{bamfile.orig}}, generating ..." 1>&2
	({{args.samtools}} index "{{bamfile | realpath}}" && \\
		ln -s "{{bamfile | realpath}}.bai" "{{bamfile}}.bai" && \\
		echo "[BioProcs] Generated: {{bamfile | realpath}}.bai" 1>&2) || \\
	({{args.samtools}} index "{{bamfile.orig}}" && \\
		ln -s "{{bamfile.orig}}.bai" "{{bamfile}}.bai" && \\
		echo "[BioProcs] Generated: {{bamfile.orig}}.bai" 1>&2) || \\
	({{args.samtools}} index "{{bamfile}}" && \\
		echo "[BioProcs] Generated: {{bamfile}}.bai" 1>&2)
fi
"""

_sTumorBai = """
if [[ ! -e "{{brings.tumor}}" ]]; then
	echo "[BioProcs] Cannot find index file for tumor: {{tumor.orig}}, generating ..." 1>&2
	({{args.samtools}} index "{{tumor | realpath}}" && \\
		ln -s "{{tumor | realpath}}.bai" "{{tumor}}.bai" && \\
		echo "[BioProcs] Generated: {{tumor | realpath}}.bai" 1>&2) || \\
	({{args.samtools}} index "{{tumor.orig}}" && \\
		ln -s "{{tumor.orig}}.bai" "{{tumor}}.bai" && \\
		echo "[BioProcs] Generated: {{tumor.orig}}.bai" 1>&2) || \\
	({{args.samtools}} index "{{tumor}}" && \\
		echo "[BioProcs] Generated: {{tumor}}.bai" 1>&2)
fi
"""

_sNormalBai = """
if [[ ! -e "{{brings.normal}}" ]]; then
	echo "[BioProcs] Cannot find index file for tumor: {{tumor.orig}}, generating ..." 1>&2
	({{args.samtools}} index "{{normal | realpath}}" && \\
		ln -s "{{normal | realpath}}.bai" "{{normal}}.bai" && \\
		echo "[BioProcs] Generated: {{normal | realpath}}.bai" 1>&2) || \\
	({{args.samtools}} index "{{normal.orig}}" && \\
		ln -s "{{normal.orig}}.bai" "{{normal}}.bai" &&
		echo "[BioProcs] Generated: {{normal.orig}}.bai" 1>&2) || \\
	({{args.samtools}} index "{{normal}}" && \\
		echo "[BioProcs] Generated: {{normal}}.bai" 1>&2)
fi
"""

@procfactory
def _pRealignerTargetCreator():
	"""
	@name:
		pRealignerTargetCreator
	@description:
		The local realignment process is designed to consume one or more BAM files and to locally realign reads such that the number of mismatching bases is minimized across all the reads. In general, a large percent of regions requiring local realignment are due to the presence of an insertion or deletion (indels) in the individual's genome with respect to the reference genome. Such alignment artifacts result in many bases mismatching the reference near the misalignment, which are easily mistaken as SNPs. Moreover, since read mapping algorithms operate on each read independently, it is impossible to place reads on the reference genome such that mismatches are minimized across all reads. Consequently, even when some reads are correctly mapped with indels, reads covering the indel near just the start or end of the read are often incorrectly mapped with respect the true indel, also requiring realignment. Local realignment serves to transform regions with misalignments due to indels into clean reads containing a consensus indel suitable for standard variant discovery approaches.
		Note that indel realignment is no longer necessary for variant discovery if you plan to use a variant caller that performs a haplotype assembly step, such as HaplotypeCaller or MuTect2. However it is still required when using legacy callers such as UnifiedGenotyper or the original MuTect. There are 2 steps to the realignment process:
		- Determining (small) suspicious intervals which are likely in need of realignment (RealignerTargetCreator)
		- Running the realigner over those intervals (see the IndelRealigner tool)
		For more details, see [the indel realignment method documentation](http://www.broadinstitute.org/gatk/guide/article?id=38).
	@input:
		`bamfile:file`:  The aligned bam file
		`reffile`: The reference file
	@brings:
		`bamfile`: `{{bamfile | bn}}.bai` The index file of input bam file
		`reffile#fai`: `{{reffile | bn}}.fai`
		`reffile#dict`: `{{reffile | bn}}.dict`
	@output:
		`outfile:file`: A list of target intervals to pass to the IndelRealigner.
	@args:
		`gatk`:     The gatk executable, default: "gatk"
		`picard`:   The picard executable, default: "picard"
		`params`:   Other parameters for RealignerTargetCreator, default: ""
		`samtools`: The samtools executable, default: "samtools"
		`tmpdir`:  The tmpdir to use. Default: /tmp
		`javamem`: The memory for java vm. Default: "-Xms1g -Xmx8g"
	@requires:
		[GATK](https://software.broadinstitute.org/gatk)
		[samtools](http://www.htslib.org/) if `reffile` is not indexed or `bamfile` is not indexed.
		[picard](https://broadinstitute.github.io/picard/) if `reffile` is not dicted.
	"""
	pRealignerTargetCreator = Proc()
	pRealignerTargetCreator.input  = "bamfile:file, reffile:file"
	# pRealignerTargetCreator.brings = {"bamfile": "{{bamfile | bn}}.bai", "reffile#dict": "{{reffile | fn}}.dict", "reffile#fai": "{{reffile | bn}}.fai"}
	pRealignerTargetCreator.output = "outfile:file:{{bamfile | fn}}.indelRealigner.intervals"
	pRealignerTargetCreator.args   = { "gatk": "gatk", "picard": "picard", "samtools": "samtools", "javamem": "-Xms1g -Xmx8g", "tmpdir": "/tmp", "params": "" }
	pRealignerTargetCreator.script = """
	%s%s%s
	# run cmd
	{{args.gatk}} -T RealignerTargetCreator {{args.javamem}} -Djava.io.tmpdir="{{args.tmpdir}}" -R "{{reffile}}" -I "{{bamfile}}" -o "{{outfile}}" {{args.params}}
	""" % (_sReferenceFai, _sReferenceDict, _sBamBai)
	return pRealignerTargetCreator

@procfactory
def _pIndelRealigner ():
	"""
	@name:
		pIndelRealigner
	@description:
		The local realignment process is designed to consume one or more BAM files and to locally realign reads such that the number of mismatching bases is minimized across all the reads. In general, a large percent of regions requiring local realignment are due to the presence of an insertion or deletion (indels) in the individual's genome with respect to the reference genome. Such alignment artifacts result in many bases mismatching the reference near the misalignment, which are easily mistaken as SNPs. Moreover, since read mapping algorithms operate on each read independently, it is impossible to place reads on the reference genome such at mismatches are minimized across all reads. Consequently, even when some reads are correctly mapped with indels, reads covering the indel near just the start or end of the read are often incorrectly mapped with respect the true indel, also requiring realignment. Local realignment serves to transform regions with misalignments due to indels into clean reads containing a consensus indel suitable for standard variant discovery approaches.
		Note that indel realignment is no longer necessary for variant discovery if you plan to use a variant caller that performs a haplotype assembly step, such as HaplotypeCaller or MuTect2. However it is still required when using legacy callers such as UnifiedGenotyper or the original MuTect.
		There are 2 steps to the realignment process:
		- Determining (small) suspicious intervals which are likely in need of realignment (see the RealignerTargetCreator tool)
		- Running the realigner over those intervals (IndelRealigner)
		For more details, see [the indel realignment method documentation](http://www.broadinstitute.org/gatk/guide/article?id=38).
	@input:
		`bamfile:file`: The aligned bam file
		`intfile:file`: Intervals file output from RealignerTargetCreator
		`reffile:file`: The reference file
	@brings:
		`bamfile`: `{{bamfile | bn}}.bai` The index file of input bam file
		`reffile#fai`: `{{reffile | bn}}.fai`
		`reffile#dict`: `{{reffile | bn}}.dict`
	@output:
		`outfile:file`: A realigned version of input BAM file.
	@args:
		`gatk`:     The gatk executable, default: "gatk"
		`picard`:   The picard executable, default: "picard"
		`params`:  Other parameters for IndelRealigner, default: ""
		`samtools`: The samtools executable, default: samtools
		`tmpdir`:  The tmpdir to use. Default: /tmp
		`javamem`: The memory for java vm. Default: "-Xms1g -Xmx8g"
	@requires:
		[GATK](https://software.broadinstitute.org/gatk)
		[samtools](http://www.htslib.org/) if `reffile` is not indexed or `bamfile` is not indexed.
		[picard](https://broadinstitute.github.io/picard/) if `reffile` is not dicted.
	"""
	pIndelRealigner = Proc()
	pIndelRealigner.input  = "bamfile:file, intfile:file, reffile:file"
	# pIndelRealigner.brings = {"bamfile": "{{bamfile | bn}}.bai", "reffile#dict": "{{reffile | fn}}.dict", "reffile#fai": "{{reffile | bn}}.fai"}
	pIndelRealigner.output = "outfile:file:{{bamfile | fn | lambda x: __import__('re').sub(r'(\\.dedup)?$', '', x)}}.realigned.bam"
	pIndelRealigner.args   = { "gatk": "gatk", "samtools": "samtools", "picard": "picard", "javamem": "-Xms1g -Xmx8g", "tmpdir": "/tmp", "params": "" }
	pIndelRealigner.script = """
	%s%s%s
	return pIndelRealigner
	{{args.gatk}} -T IndelRealigner {{args.javamem}} -Djava.io.tmpdir="{{args.tmpdir}}" -R "{{reffile}}" -I "{{bamfile}}" -o "{{outfile}}" -targetIntervals "{{intfile}}" {{args.params}}
	""" % (_sReferenceFai, _sReferenceDict, _sBamBai)
	return pIndelRealigner

@procfactory
def _pBaseRecalibrator():
	"""
	@name:
		pBaseRecalibrator
	@description:
		Variant calling algorithms rely heavily on the quality scores assigned to the individual base calls in each sequence read. These scores are per-base estimates of error emitted by the sequencing machines. Unfortunately the scores produced by the machines are subject to various sources of systematic technical error, leading to over- or under-estimated base quality scores in the data. Base quality score recalibration (BQSR) is a process in which we apply machine learning to model these errors empirically and adjust the quality scores accordingly. This allows us to get more accurate base qualities, which in turn improves the accuracy of our variant calls. The base recalibration process involves two key steps: first the program builds a model of covariation based on the data and a set of known variants (which you can bootstrap if there is none available for your organism), then it adjusts the base quality scores in the data based on the model. There is an optional but highly recommended step that involves building a second model and generating before/after plots to visualize the effects of the recalibration process. This is useful for quality control purposes. This tool performs the first step described above: it builds the model of covariation and produces the recalibration table. It operates only at sites that are not in dbSNP; we assume that all reference mismatches we see are therefore errors and indicative of poor base quality. This tool generates tables based on various user-specified covariates (such as read group, reported quality score, cycle, and context). Assuming we are working with a large amount of data, we can then calculate an empirical probability of error given the particular covariates seen at this site, where p(error) = num mismatches / num observations. The output file is a table (of the several covariate values, number of observations, number of mismatches, empirical quality score).
	@input:
		`bamfile:file`: A BAM file containing data that needs to be recalibrated.
		`reffile:file`: The reference file
	@brings:
		`bamfile`: `{{bamfile | bn}}.bai` The index file of input bam file
		`reffile#fai`: `{{reffile | bn}}.fai`
		`reffile#dict`: `{{reffile | bn}}.dict`
	@output:
		`outfile:file`: A GATKReport file with many tables:
			- The list of arguments
			- The quantized qualities table
			- The recalibration table by read group
			- The recalibration table by quality score
			- The recalibration table for all the optional covariates
	@args:
		`gatk`:     The gatk executable, default: "gatk"
		`params`:  Other parameters for BaseRecalibrator, default: ""
		`knownSites`: The known polymorphic sites to mask out, required
		`samtools`: The samtools executable, default: samtools
		`picard`:   The picard executable, default: "picard"
		`tmpdir`:  The tmpdir to use. Default: /tmp
		`javamem`: The memory for java vm. Default: "-Xms1g -Xmx8g"
	@requires:
		[GATK](https://software.broadinstitute.org/gatk)
		[samtools](http://www.htslib.org/) if `reffile` is not indexed or `bamfile` is not indexed.
		[picard](https://broadinstitute.github.io/picard/) if `reffile` is not dicted.
	"""
	pBaseRecalibrator = Proc()
	pBaseRecalibrator.input  = "bamfile:file, reffile:file"
	# pBaseRecalibrator.brings = {"bamfile": "{{bamfile | bn}}.bai", "reffile#dict": "{{reffile | fn}}.dict", "reffile#fai": "{{reffile | bn}}.fai"}
	pBaseRecalibrator.output = "outfile:file:{{bamfile | fn}}.recal.table"
	pBaseRecalibrator.args   = { "gatk": "gatk", "samtools": "samtools", "picard": "picard", "javamem": "-Xms1g -Xmx8g", "tmpdir": "/tmp", "knownSites": "", "params": "" }
	pBaseRecalibrator.script = """
	if [[ -z "{{args.knownSites}}" ]]; then
		echo "knownSites file not specified!" 1>&2
		exit 1
	fi
	%s%s%s
	{{args.gatk}} -T BaseRecalibrator {{args.javamem}} -Djava.io.tmpdir="{{args.tmpdir}}" -R "{{reffile}}" -I "{{bamfile}}" -o "{{outfile}}" -knownSites "{{args.knownSites}}" {{args.params}}
	""" % (_sReferenceFai, _sReferenceDict, _sBamBai)
	return pBaseRecalibrator

@procfactory
def _pPrintReads():
	"""
	@name:
		pPrintReads
	@description:
		PrintReads is a generic utility tool for manipulating sequencing data in SAM/BAM format. It can dynamically merge the contents of multiple input BAM files, resulting in merged output sorted in coordinate order. It can also optionally filter reads based on various read properties such as read group tags using the `--read_filter/-rf` command line argument (see documentation on read filters for more information).
		Note that when PrintReads is used as part of the Base Quality Score Recalibration workflow, it takes the `--BQSR` engine argument, which is listed under Inherited Arguments > CommandLineGATK below.
	@input:
		`bamfile:file`:    A BAM file.
		`recaltable:file`: The GATKReport file
		`reffile:file`:    The reference file
	@brings:
		`bamfile`: `{{bamfile | bn}}.bai` The index file of input bam file
		`reffile#fai`: `{{reffile | bn}}.fai`
		`reffile#dict`: `{{reffile | bn}}.dict`
	@output:
		`outfile:file`: A single processed bam file.
	@args:
		`gatk`:     The gatk executable, default: "gatk"
		`params`:  Other parameters for PrintReads, default: ""
		`samtools`: The samtools executable, default: samtools
		`picard`:   The picard executable, default: "picard"
		`tmpdir`:  The tmpdir to use. Default: /tmp
		`javamem`: The memory for java vm. Default: "-Xms1g -Xmx8g"
	@requires:
		[GATK](https://software.broadinstitute.org/gatk)
		[samtools](http://www.htslib.org/) if `reffile` is not indexed or `infile` is not indexed.
		[picard](https://broadinstitute.github.io/picard/) if `reffile` is not dicted.
	"""
	pPrintReads = Proc()
	pPrintReads.input  = "bamfile:file, recaltable:file, reffile:file"
	# pPrintReads.brings = {"bamfile": "{{bamfile | bn}}.bai", "reffile#dict": "{{reffile | fn}}.dict", "reffile#fai": "{{reffile | bn}}.fai"}
	pPrintReads.output = "outfile:file:{{bamfile | fn | lambda x: __import__('re').sub(r'(\\.realigned)?$', '', x)}}.bam"
	pPrintReads.args   = { "gatk": "gatk", "samtools": "samtools", "picard": "picard", "javamem": "-Xms1g -Xmx8g", "tmpdir": "/tmp", "params": "" }
	pPrintReads.script = """
	%s%s%s
	{{args.gatk}} -T PrintReads {{args.javamem}} -Djava.io.tmpdir="{{args.tmpdir}}" -R "{{reffile}}" -I "{{bamfile}}" -o "{{outfile}}" -BQSR "{{recaltable}}" {{args.params}}
	""" % (_sReferenceFai, _sReferenceDict, _sBamBai)
	return pPrintReads

@procfactory
def _pHaplotypeCaller():
	"""
	@name:
		pHaplotypeCaller
	@description:
		PrintReads is a generic utility tool for manipulating sequencing data in SAM/BAM format. It can dynamically merge the contents of multiple input BAM files, resulting in merged output sorted in coordinate order. It can also optionally filter reads based on various read properties such as read group tags using the `--read_filter/-rf` command line argument (see documentation on read filters for more information).
		Note that when PrintReads is used as part of the Base Quality Score Recalibration workflow, it takes the `--BQSR` engine argument, which is listed under Inherited Arguments > CommandLineGATK below.
	@input:
		`bamfile:file`: A BAM file.
		`reffile:file`: The reference file
	@brings:
		`bamfile`: `{{bamfile | bn}}.bai` The index file of input bam file
		`reffile#fai`: `{{reffile | bn}}.fai`
		`reffile#dict`: `{{reffile | bn}}.dict`
	@output:
		`outfile:file`: Either a VCF or gVCF file with raw, unfiltered SNP and indel calls.
	@args:
		`gatk`    : The gatk executable, default: "gatk"
		`params`  : Other parameters for HaplotypeCaller, default: ""
		`samtools`: The samtools executable, default: samtools
		`picard`:   The picard executable, default: "picard"
		`tmpdir`:  The tmpdir to use. Default: /tmp
		`javamem`: The memory for java vm. Default: "-Xms1g -Xmx8g"
		`nthread`: Corresponding to -nct option
	@requires:
		[GATK](https://software.broadinstitute.org/gatk)
		[samtools](http://www.htslib.org/) if `reffile` is not indexed or `infile` is not indexed.
		[picard](https://broadinstitute.github.io/picard/) if `reffile` is not dicted.
	"""
	pHaplotypeCaller = Proc()
	pHaplotypeCaller.input  = "bamfile:file, reffile:file"
	# pHaplotypeCaller.brings = {"bamfile": "{{bamfile | bn}}.bai", "reffile#dict": "{{reffile | fn}}.dict", "reffile#fai": "{{reffile | bn}}.fai"}
	pHaplotypeCaller.output = "outfile:file:{{bamfile | fn}}.vcf"
	pHaplotypeCaller.args   = { "gatk": "gatk", "samtools": "samtools", "picard": "picard", "javamem": "-Xms1g -Xmx8g", "tmpdir": "/tmp", "nthread": 1, "params": "" }
	pHaplotypeCaller.script = """
	%s%s%s
	{{args.gatk}} -T HaplotypeCaller {{args.javamem}} -Djava.io.tmpdir="{{args.tmpdir}}" -R "{{reffile}}" -I "{{bamfile}}" -o "{{outfile}}" -nct {{args.nthread}} {{args.params}}
	""" % (_sReferenceFai, _sReferenceDict, _sBamBai)
	return pHaplotypeCaller

@procfactory
def _pSelectVariants():
	"""
	@name:
		pSelectVariants
	@description:
		Often, a VCF containing many samples and/or variants will need to be subset in order to facilitate certain analyses (e.g. comparing and contrasting cases vs. controls; extracting variant or non-variant loci that meet certain requirements, displaying just a few samples in a browser like IGV, etc.). SelectVariants can be used for this purpose.
		There are many different options for selecting subsets of variants from a larger callset:
		- Extract one or more samples from a callset based on either a complete sample name or a pattern match.
		- Specify criteria for inclusion that place thresholds on annotation values, e.g. "DP > 1000" (depth of coverage greater than 1000x), "AF < 0.25" (sites with allele frequency less than 0.25). These - criteria are written as "JEXL expressions", which are documented in the article about using JEXL expressions.
		- Provide concordance or discordance tracks in order to include or exclude variants that are also present in other given callsets.
		- Select variants based on criteria like their type (e.g. INDELs only), evidence of mendelian violation, filtering status, allelicity, and so on.
		There are also several options for recording the original values of certain annotations that are recalculated when a subsetting the new callset, trimming alleles, and so on.
	@input:
		`vcffile:file`: A variant call set from which to select a subset.
		`reffile:file`: The reference file
	@brings:
		`reffile#fai`: `{{reffile | bn}}.fai`
		`reffile#dict`: `{{reffile | bn}}.dict`
	@output:
		`outfile:file`: A new VCF file containing the selected subset of variants.
	@args:
		`gatk`:     The gatk executable, default: "gatk"
		`params`:  Other parameters for SelectVariants, default: ""
		`samtools`: The samtools executable, default: samtools
		`picard`:   The picard executable, default: "picard"
		`tmpdir`:  The tmpdir to use. Default: /tmp
		`javamem`: The memory for java vm. Default: "-Xms1g -Xmx8g"
	@requires:
		[GATK](https://software.broadinstitute.org/gatk)
		[samtools](http://www.htslib.org/) if `reffile` is not indexed or `infile` is not indexed.
		[picard](https://broadinstitute.github.io/picard/) if `reffile` is not dicted.
	"""
	pSelectVariants = Proc()
	pSelectVariants.input  = "vcffile:file, reffile:file"
	# pSelectVariants.brings = {"reffile#fai": "{{reffile | bn}}.fai", "reffile#dict": "{{reffile | fn}}.dict"}
	pSelectVariants.output = "outfile:file:{{vcffile | fn}}.selected.vcf"
	pSelectVariants.args   = {"gatk": "gatk", "samtools": "samtools", "picard": "picard", "javamem": "-Xms1g -Xmx8g", "tmpdir": "/tmp", "params": ""}
	pSelectVariants.script = """
	%s%s
	{{args.gatk}} -T SelectVariants {{args.javamem}} -Djava.io.tmpdir="{{args.tmpdir}}" -R "{{reffile}}" -V "{{vcffile}}" -o "{{outfile}}" {{args.params}}
	""" % (_sReferenceFai, _sReferenceDict)
	return pSelectVariants

@procfactory
def _pVariantFiltration():
	"""
	@name:
		pVariantFiltration
	@description:
		This tool is designed for hard-filtering variant calls based on certain criteria. Records are hard-filtered by changing the value in the FILTER field to something other than PASS. Filtered records will be preserved in the output unless their removal is requested in the command line.
		The most common way of specifying filtering criteria is by using JEXL queries. See the article on JEXL expressions in the documentation Guide for detailed information and examples.
	@input:
		`vcffile:file`: A variant call set from which to select a subset.
		`reffile:file`: The reference file
	@brings:
		`reffile#fai`: `{{reffile | bn}}.fai`
		`reffile#dict`: `{{reffile | bn}}.dict`
	@output:
		`outfile:file`: A filtered VCF.
	@args:
		`gatk`:     The gatk executable, default: "gatk -T VariantFiltration"
		`params`:  Other parameters for VariantFiltration, default: ""
		`samtools`: The samtools executable, default: samtools
		`picard`:   The picard executable, default: "picard"
		`tmpdir`:  The tmpdir to use. Default: /tmp
		`javamem`: The memory for java vm. Default: "-Xms1g -Xmx8g"
	@requires:
		[GATK](https://software.broadinstitute.org/gatk)
		[samtools](http://www.htslib.org/) if `reffile` is not indexed or `infile` is not indexed.
		[picard](https://broadinstitute.github.io/picard/) if `reffile` is not dicted.
	"""
	pVariantFiltration = Proc()
	pVariantFiltration.input  = "vcffile:file, reffile:file"
	# pVariantFiltration.brings = {"reffile#fai": "{{reffile | bn}}.fai", "reffile#dict": "{{reffile | fn}}.dict"}
	pVariantFiltration.output = "outfile:file:{{vcffile | fn | lambda x: __import__('re').sub(r'(\\.selected)?$', '', x)}}.filtered.vcf"
	pVariantFiltration.args   = { "gatk": "gatk", "samtools": "samtools", "picard": "picard", "javamem": "-Xms1g -Xmx8g", "tmpdir": "/tmp", "params": "" }
	pVariantFiltration.script = """
	%s%s
	{{args.gatk}} -T VariantFiltration {{args.javamem}} -Djava.io.tmpdir="{{args.tmpdir}}" -R "{{reffile}}" -V "{{vcffile}}" -o "{{outfile}}" {{args.params}}
	""" % (_sReferenceFai, _sReferenceDict)
	return pVariantFiltration

@procfactory
def _pMuTect2():
	"""
	@name:
		pMuTect2
	@description:
		MuTect2 is a somatic SNP and indel caller that combines the DREAM challenge-winning somatic genotyping engine of the original MuTect ([Cibulskis et al., 2013](http://www.nature.com/nbt/journal/v31/n3/full/nbt.2514.html)) with the assembly-based machinery of HaplotypeCaller. The basic operation of MuTect2 proceeds similarly to that of the HaplotypeCaller.
		NOTE: only Tumor/Normal variant calling implemented in bioprocs
	@input:
		`tumor:file`:   the tumor bam file
		`normal:file`:  the normal bam file
		`reffile:file`: the reference file
	@brings:
		`tumor`:  `{{tumor | bn}}.bai` the index file of tumor
		`normal`: `{{normal | bn}}.bai` the index file of normal
		`reffile#fai`: `{{reffile | bn}}.fai`
		`reffile#dict`: `{{reffile | bn}}.dict`
	@output:
		`outfile:file`: The vcf file containing somatic mutations
	@args:
		`gatk`:     The gatk executable, default: "gatk"
		`samtools`: The samtools executable, default: samtools
		`params`:   Other parameters for MuTect2, default: ""
		`picard`:   The picard executable, default: "picard"
		`tmpdir`:  The tmpdir to use. Default: /tmp
		`javamem`: The memory for java vm. Default: "-Xms1g -Xmx8g"
	@requires:
		[GATK](https://software.broadinstitute.org/gatk)
		[samtools](http://www.htslib.org/) if index files of input files are not found
		[picard](https://broadinstitute.github.io/picard/) if `reffile` is not dicted.
	"""
	pMuTect2 = Proc()
	pMuTect2.input  = "tumor:file, normal:file, reffile:file"
	#pMuTect2.brings = {"tumor": "{{tumor | bn}}.bai", "normal": "{{normal | bn}}.bai", "reffile#fai": "{{reffile | bn}}.fai", "reffile#dict": "{{reffile | fn}}.dict"}
	pMuTect2.output = "outfile:file:{{tumor | fn}}-{{normal | fn}}.vcf"
	pMuTect2.args   = { "gatk": "gatk", "samtools": "samtools", "picard": "picard", "javamem": "-Xms1g -Xmx8g", "tmpdir": "/tmp", "params": "" }
	pMuTect2.script = """
	%s%s%s%s
	{{args.gatk}} -T MuTect2 {{args.javamem}} -Djava.io.tmpdir="{{args.tmpdir}}" -R "{{reffile}}" -I:tumor "{{tumor}}" -I:normal "{{normal}}" -o "{{outfile}}" {{args.params}}
	""" % (_sNormalBai, _sTumorBai, _sReferenceFai, _sReferenceDict)
	return pMuTect2

@procfactory
def _pMuTect2Interval():
	"""
	@name:
		pMuTect2Interval
	@description:
		Use interval file model of MuTect2
	@input:
		`tumor:file`:   the tumor bam file
		`normal:file`:  the normal bam file
		`reffile:file`: the reference file
	@brings:
		`tumor`:  `{{tumor | bn}}.bai` the index file of tumor
		`normal`: `{{normal | bn}}.bai` the index file of normal
		`reffile#fai`: `{{reffile | bn}}.fai`
		`reffile#dict`: `{{reffile | bn}}.dict`
	@output:
		`outfile:file`: The vcf file containing somatic mutations
	@args:
		`gatk`:     The gatk executable, default: "gatk"
		`samtools`: The samtools executable, default: samtools
		`params`:   Other parameters for MuTect2, default: ""
		`picard`:   The picard executable, default: "picard"
		`tmpdir`:  The tmpdir to use. Default: /tmp
		`javamem`: The memory for java vm. Default: "-Xms1g -Xmx8g"
	@requires:
		[GATK](https://software.broadinstitute.org/gatk)
		[samtools](http://www.htslib.org/) if index files of input files are not found
		[picard](https://broadinstitute.github.io/picard/) if `reffile` is not dicted.
	"""
	pMuTect2Interval = Proc()
	pMuTect2Interval.input  = "tumor:file, normal:file, reffile:file"
	#pMuTect2Interval.brings = {"tumor": "{{tumor | bn}}.bai", "normal": "{{normal | bn}}.bai", "reffile#fai": "{{reffile | bn}}.fai", "reffile#dict": "{{reffile | fn}}.dict"}
	pMuTect2Interval.output = "outfile:file:{{tumor | fn}}-{{normal | fn}}.vcf"
	pMuTect2Interval.args   = { "gatk": "gatk", "picard": "picard", "samtools": "samtools", "javamem": "-Xms1g -Xmx8g", "tmpdir": "/tmp", "params": "" }
	pMuTect2Interval.script = """
	%s%s%s%s
	# generate interval file
	intvfile="{{job.outdir}}/interval.list"
	{{args.samtools}} idxstats "{{tumor}}" | head -n -1 | cut -f1 > $intvfile
	{{args.gatk}} -T MuTect2 {{args.javamem}} -Djava.io.tmpdir="{{args.tmpdir}}" -R "{{reffile}}" -I:tumor "{{tumor}}" -I:normal "{{normal}}" -o "{{outfile}}" -L $intvfile {{args.params}}
	""" % (_sNormalBai, _sTumorBai, _sReferenceFai, _sReferenceDict)
	return pMuTect2Interval




from pyppl import Box
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

files   = {{i.infiles | repr}}
inopts  = {{args.inopts | repr}}
outopts = {{args.outopts | repr}}
outfile = {{o.outfile | quote}}
maxopen = {{args.maxopen | repr}}

def getReaders(fs, baseidx):
	readers = []
	for i,infile in enumerate(fs):
		reader = TsvReader(infile, **inopts)
		if not reader.meta:
			reader.autoMeta()
		readers.append(reader)
	return readers

writer = TsvWriter(outfile, delimit = inopts['delimit'])
if inopts.get('cnames', True):
	reader = TsvReader(files[0], **inopts)
	reader.close()
	writer.cnames = reader.cnames
	writer.writeHead()

for i in range(0, len(files), maxopen):
	fs = files[i:i + maxopen]
	readers = getReaders(fs, i)

	for reader in readers:
		for r in reader:
			writer.write(r)
		reader.close()
writer.close()



__version__ = '0.1.2'

import inspect
from pathlib import Path
from tempfile import gettempdir
from sys import executable, modules
from pyppl import Proc
from pyparam import Params

# open to R (reticulate) to get the path of r util scripts
HERE    = Path(__file__).resolve().parent
DEFAULTS = {
	# constants
	"dbsnpver"        : "150",
	"dbsnpver.desc"   : "The dbsnp version used in cruzdb to query SNP information.",
	"genome"          : "hg19",
	"genome.desc"     : "Commonly used genome assembly.",
	"ncbikey"         : "",
	"ncbikey.desc"    : "The NCBI API Key for E-Utils.",
	"chrorder"        : ",".join('chr' + str(x) for x in list(range(1,23)) + ['X', 'Y', 'M', 'MT']),
	"chrorder.desc"   : "Preferrable chromsome orders, used for sorting.",

	# path
	"affysnps"           : "",
	"affysnps.desc"      : "Affymetrix SNP 6.0 Positions, used for THetA2 mostly.",
	"annovarDb"          : "",
	"annovarDb.desc"     : "The path of database for Annovar.",
	"cachedir"           : str(Path.home() / 'cache'),
	"cachedir.desc"      : "The directory to cache query data.",
	"consvdir"           : "",
	"consvdir.desc"      : "The directory containing conservation scores in bigWig files.\nUse ucsc-wig2bigwig the original files are wigFix files.",
	"cytoband"           : "",
	"cytoband.desc"      : "The cytoband file from ucsc database.",
	"dbsnp"              : "",
	"dbsnp.desc"         : "The dbsnp common variants in VCF format.",
	"dbsnp_all"          : "",
	"dbsnp_all.desc"     : "The dbsnp all variants in VCF format.",
	"gep70"              : "",
	"gep70.desc"         : "The GEP70 genes (col1: 51up, col2:19down).",
	"gsize"              : "",
	"gsize.desc"         : "The chromsome size file.",
	"highld"             : "",
	"highld.desc"        : "High LD regions",
	"ipidb"              : "",
	"ipidb.desc"         : "The IPI xref database file.",
	"kallistoIdx"        : "",
	"kallistoIdx.desc"   : "The kallisto index file.",
	"lcsig"              : "",
	"lcsig.desc"         : "Lung cancer gene signature (see https://www.ncbi.nlm.nih.gov/pubmed/19118056)",
	"liftover"           : "liftOver",
	"liftover.desc"      : "The liftover binary file.",
	"lochain"            : "",
	"lochain.desc"       : "The liftover chain file.",
	"mcr"                : "",
	"mcr.desc"           : "Matlab MCR path.",
	"mutsig_cvrg"        : "",
	"mutsig_cvrg.desc"   : "Coverage table for MutSig.",
	"mutsig_cvrt"        : "",
	"mutsig_cvrt.desc"   : "Covariates table for MutSig.",
	"mutsig_mutdict"     : "",
	"mutsig_mutdict.desc": "Mutation type dictionary file for MutSig.",
	"mutsig_chrdir"      : "",
	"mutsig_chrdir.desc" : "Directory containing reference sequence files for MutSig.",
	"oncotator_db"       : "",
	"oncotator_db.desc"  : "The data directory for oncotator.",
	"ref"                : "",
	"ref.desc"           : "The reference genome.",
	"refgene"            : "",
	"refgene.desc"       : "The reference genes in GTF format.",
	"refexon"            : "",
	"refexon.desc"       : "The exons of reference genes in GTF format.",
	"rsmap_gwsnp6"       : "",
	"rsmap_gwsnp6.desc"  : "GenomeWideSNP_6 RSID SNP probe mapping file.",
	"snpeffDb"           : "",
	"snpeffDb.desc"      : "The path of database for snpEff.",
	"superfreq_res"      : "",
	"superfreq_res.desc" : "superFreq resource directory.",
	"tflist"             : "",
	"tflist.desc"        : "The TF list file with motif as 1st column and TF name as 2nd column.",
	"tfmotifs"           : "",
	"tfmotifs.desc"      : "The TF motifs in MEME format.",
	"tmpdir"             : gettempdir(),
	"tmpdir.desc"        : "The temporary directory.",
	"vepNonTCGAVcf"      : "",
	"vepNonTCGAVcf.desc" : "The path of ExAC_nonTCGA.r0.3.1.sites.vep.vcf.gz for VEP.",
	"vepDb"              : "",
	"vepDb.desc"         : "The path of database for VEP.",


	# memoreis
	"mem4G"      : "4G",
	"mem4G.desc" : "4GB memory.",
	"mem8G"      : "8G",
	"mem8G.desc" : "8GB memory.",
	"mem16G"     : "16G",
	"mem16G.desc": "16GB memory.",
	"mem24G"     : "24G",
	"mem24G.desc": "24GB memory.",
	"mem32G"     : "32G",
	"mem32G.desc": "32GB memory.",

	# tools
	"allfit"                   : "All-FIT.py",
	"allfit.desc"              : "Path to All-FIT.py",
	"annovar"                  : "annotate_variation.pl",
	"annovar.desc"             : "The path of annovar's annotate_variation.pl.",
	"annovar_convert"          : "convert2annovar.pl",
	"annovar_convert.desc"     : "The path of annovar's convert2annovar.pl.",
	"awk"                      : "awk",
	"awk.desc"                 : "The awk program.",
	"arsample"                 : "sample",
	"arsample.desc"            : "The tool `sample` by Alex Reynolds.",
	"bam_readcount"            : "bam-readcount",
	"bam_readcount.desc"       : "Program to generate metrics at single nucleotide positions",
	"bamstats"                 : "bamstats",
	"bamstats.desc"            : "The path of bamstats.",
	"bamutil"                  : "bam",
	"bamutil.desc"             : "The path of bamutil.",
	"bcftools"                 : "bcftools",
	"bcftools.desc"            : "The path of bcftools.",
	"bedops_sort"              : "sort-bed",
	"bedops_sort.desc"         : "The path of sort-bed of bedops.",
	"bedtools"                 : "bedtools",
	"bedtools.desc"            : "The path of bedtools.",
	"biobambam_bamsort"        : "bamsort",
	"biobambam_bamsort.desc"   : "The path of bamsort of biobambam.",
	"biobambam_bamtofastq"     : "bamtofastq",
	"biobambam_bamtofastq.desc": "The path of bamtofastq of biobambam.",
	"bowtie2"                  : "bowtie2",
	"bowtie2.desc"             : "The path of bamtofastq of bowtie2.",
	"bwa"                      : "bwa",
	"bwa.desc"                 : "The path of bamtofastq of bwa.",
	"bwtool"                   : "bwtool",
	"bwtool.desc"              : "The path of bwtool.",
	"cnvkit"                   : "cnvkit.py",
	"cnvkit.desc"              : "The path of cnvkit.py.",
	"cnvnator"                 : "cnvnator",
	"cnvnator.desc"            : "The path of cnvnator.",
	"cnvnator2vcf"             : "cnvnator2VCF.pl",
	"cnvnator2vcf.desc"        : "The path of cnvnator2VCF.pl.",
	"curl"                     : "curl",
	"curl.desc"                : "The path of command line tool curl.",
	"cutadapt"                 : "cutadapt",
	"cutadapt.desc"            : "The path of cutadapt.",
	"dot"                      : "dot",
	"dot.desc"                 : "The path of dot.",
	"dtoxog"                   : "dtoxog",
	"dtoxog.desc"              : "The path of dtoxog, a wrapper of D-ToxoG.",
	"dwgsim"                   : "dwgsim",
	"dwgsim.desc"              : "The path of dwgsim.",
	"elprep"                   : "elprep",
	"elprep.desc"              : "The path of elprep.",
	"fastqc"                   : "fastqc",
	"fastqc.desc"              : "The path of fastqc.",
	"fimo"                     : "fimo",
	"fimo.desc"                : "The path of fimo from MEME suite.",
	"gatk"                     : "gatk",
	"gatk.desc"                : "The path of gatk.",
	"gdc_client"               : "gdc-client",
	"gdc_client.desc"          : "The GDC client for TCGA data download.",
	"gistic"                   : "gistic2",
	"gistic.desc"              : "The path of gistic.",
	"htseq_count"              : "htseq-count",
	"htseq_count.desc"         : "The path of htseq-count.",
	"lichee"                   : "lichee",
	"lichee.desc"              : "The path of lichee.",
	"jvarkit"                  : "jvarkit",
	"jvarkit.desc"             : "The path of jvarkit.",
	"kallisto"                 : "kallisto",
	"kallisto.desc"            : "The path of kallisto.",
	"maf2vcf"                  : "maf2vcf.pl",
	"maf2vcf.desc"             : "The path of maf2vcf.pl.",
	"multiqc"                  : "multiqc",
	"multiqc.desc"             : "The path of multiqc.",
	"mutsig"                   : "mutsig",
	"mutsig.desc"              : "The path of run_MutSigCV.sh",
	"netmhc"                   : "netMHC",
	"netmhc.desc"              : "The path of netMHC.",
	"netmhcpan"                : "netMHCpan",
	"netmhcpan.desc"           : "The path of netMHCpan.",
	"netmhciipan"              : "netMHCIIpan",
	"netmhciipan.desc"         : "The path of netMHCIIpan.",
	"netmhccons"               : "netMHCCons",
	"netmhccons.desc"          : "The path of netMHCCons.",
	"ngm"                      : "ngm",
	"ngm.desc"                 : "The path of ngm.",
	"oncotator"                : "oncotator",
	"oncotator.desc"           : "The path of oncotator.",
	"picard"                   : "picard",
	"picard.desc"              : "The path of picard.",
	"platypus"                 : "platypus",
	"platypus.desc"            : "The path of platypus.",
	"plink"                    : "plink",
	"plink.desc"               : "Executable of plink.",
	"pyclone"                  : "PyClone",
	"pyclone.desc"             : "The path of PyClone",
	"sambamba"                 : "sambamba",
	"sambamba.desc"            : "The path of sambamba.",
	"samtools"                 : "samtools",
	"samtools.desc"            : "The path of samtools.",
	"schism"                   : "runSchism",
	"schism.desc"              : "The path of runSchism.",
	"skewer"                   : "skewer",
	"skewer.desc"              : "The path of skewer.",
	"snpeff"                   : "snpEff",
	"snpeff.desc"              : "The path of snpEff.",
	"snpsift"                  : "SnpSift",
	"snpsift.desc"             : "The path of SnpSift.",
	"snvsniffer"               : "SNVSniffer",
	"snvsniffer.desc"          : "The path of SNVSniffer.",
	"somaticsniper"            : "bam-somaticsniper",
	"somaticsniper.desc"       : "The path of bam-somaticsniper.",
	"star"                     : "STAR",
	"star.desc"                : "The path of star.",
	"strelka_germ"             : "configureStrelkaGermlineWorkflow.py",
	"strelka_germ.desc"        : "The path of configureStrelkaGermlineWorkflow.py.",
	"strelka_soma"             : "configureStrelkaSomaticWorkflow.py",
	"strelka_soma.desc"        : "The path of configureStrelkaSomaticWorkflow.py.",
	"tabix"                    : "tabix",
	"tabix.desc"               : "The path of tabix.",
	"theta2"                   : "RunTHetA.py",
	"theta2.desc"              : "The path of THetA2 executable.",
	"tomtom"                   : "tomtom",
	"tomtom.desc"              : "tomtom from MEME suite.",
	"topiary"                  : "topiary",
	"topiary.desc"             : "The path to topiary.",
	"trimmomatic"              : "trimmomatic",
	"trimmomatic.desc"         : "The path of trimmomatic.",
	"vardict"                  : "vardict",
	"vardict.desc"             : "The path of vardict.",
	"vcf2maf"                  : "vcf2maf.pl",
	"vcf2maf.desc"             : "The path of vcf2maf.pl.",
	"vcfanno"                  : "vcfanno",
	"vcfanno.desc"             : "The path of vcfanno.",
	"vcflib_vcffilter"         : "vcffilter",
	"vcflib_vcffilter.desc"    : "The path of vcflib vcffilter.",
	"vcfstats"                 : "vcfstats",
	"vcfstats.desc"            : "The path of vcfstats.",
	"vcftools"                 : "vcftools",
	"vcftools.desc"            : "The path of vcftools.",
	"vcftools_merge"           : "vcf-merge",
	"vcftools_merge.desc"      : "The path of vcftools' vcf-merge.",
	"vcftools_subset"          : "vcf-subset",
	"vcftools_subset.desc"     : "The path of vcftools' vcf-subset.",
	"vep"                      : "vep",
	"vep.desc"                 : "The path of vep.",
	"virmid"                   : "virmid",
	"virmid.desc"              : "The path of virmid.",
	"wgsim"                    : "wgsim",
	"wgsim.desc"               : "The path of wgsim.",
	"wandy"                    : "Wandy",
	"wandy.desc"               : "The path of Wandy.",

	# langs
	"python"      : "python",
	"python.desc" : "The path of python.",
	"Rscript"     : "Rscript",
	"Rscript.desc": "The path of Rscript.",
	"bash"        : "bash",
	"bash.desc"   : "The path of bash.",
}

params = Params()
params._load(DEFAULTS)
cfgfiles = [
	Path.home() / '.bioprocs.config', # values overwritten
	Path.home() / '.bioprocs.json',
]
for cfgfile in cfgfiles:
	if not cfgfile.exists():
		continue
	params._loadFile (cfgfile)

# lock the params in case the options are overwritten unintentionally.
params._locked = True

cachedir = Path(params.cachedir.value)
if not cachedir.exists():
	cachedir.mkdir()

rimport  = """
(function(...) {
	reticulate::use_python('%s', required = TRUE)
	bioprocs = reticulate::import('bioprocs')
	for (rfile in list(...)) {
		source(file.path(bioprocs$HERE, 'utils', rfile))
	}
})""" % executable

bashimport = """
function __bashimport__ () {
	for src in "$@"; do
		source "%s/utils/$src"
	done
}
__python__='%s'
__bashimport__""" % (HERE, executable)


EXT_MAP = {
	'Rscript': 'R',
	'python' : 'py',
	'python2': 'py',
	'python3': 'py',
}
FACTORY_CACHE = {}

def delefactory():
	"""The factory to give the delegator for modkit"""
	frame  = inspect.currentframe().f_back
	module = inspect.getmodule(frame)
	def delegator(proc):
		try:
			procfac =  module._mkenvs['_' + proc]
			if not procfac:
				raise KeyError
		except KeyError as exc:
			raise ImportError('No such process: {!r}'.format(proc)) from exc
		if not callable(procfac):
			raise ImportError('Wrong type of process factory: {!r} in module {!r}'.format(
				'_' + proc, module.__name__))
		return procfac()
	return delegator

def _procfactory(procfunc, pid, alias, mdname, doc):
	def factory():
		proc = procfunc()
		if isinstance(proc, dict):
			proc = Proc(**proc)
		proc.id = alias
		proc.props.origin = pid
		lang = Path(proc.lang).name
		ext  = '.' + EXT_MAP.get(lang, lang)
		if ext == '.R' and not proc.envs.get('rimport'):
			proc.envs.rimport = rimport
		if ext == '.bash' and not proc.envs.get('bashimport'):
			proc.envs.bashimport = bashimport
		script = HERE / 'scripts' / mdname / (pid + ext)
		if not proc.config.script and script.exists():
			proc.script = 'file:%s' % script
		report = HERE / 'reports' / mdname / (pid + '.md')
		if not proc.config.report and report.exists():
			proc.report = 'file:%s' % report
		return proc
	factory.__doc__ = doc
	return factory

def procfactory(procfunc):
	mdname = procfunc.__module__.split('.')[-1]
	pid    = procfunc.__name__.lstrip('_')
	module = modules[procfunc.__module__]
	args   = inspect.signature(procfunc).parameters
	alias  = args.get('alias')
	if alias:
		alias = alias.default
		if alias[0] == '_':
			alias = alias[1:]
		module._mkenvs['_' + alias] = _procfactory(procfunc, pid, alias, mdname, procfunc.__doc__)
	return _procfactory(procfunc, pid, pid, mdname, procfunc.__doc__)



"""Utilities from bedtools"""
from pyppl import Proc, Box
from . import params, bashimport
from .utils import fs2name
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory()).exports('_p*')

@procfactory
def _pBedGetfasta():
	"""
	@name:
		pBedGetfasta
	@description:
		`bedtools getfasta` extracts sequences from a FASTA file for each of the intervals defined in a BED file.
	@input:
		`infile:file`: The input bed file
	@output:
		`outfile:file`: The generated fasta file
	@args:
		`ref`     : The fasta file
		`bedtools`: The bedtools executable,                  default: `<params.bedtools>`
		`params`  : Other parameters for `bedtools getfasta`, default: ""
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedGetfasta                 = Proc(desc = '`bedtools getfasta` extracts sequences from a FASTA file for each of the intervals defined in a BED file.')
	pBedGetfasta.input           = "infile:file"
	pBedGetfasta.output          = "outfile:file:{{i.infile | fn}}.fa"
	pBedGetfasta.args.samtools   = params.samtools.value
	pBedGetfasta.args.bedtools   = params.bedtools.value
	pBedGetfasta.args.params     = Box(name = True)
	pBedGetfasta.args.ref        = params.ref.value
	pBedGetfasta.envs.bashimport = bashimport
	pBedGetfasta.beforeCmd       = '''
	{{bashimport}} reference.bash
	export samtools={{args.samtools | squote}}
	reference fasta {{args.ref | squote}}
	'''
	pBedGetfasta.lang   = params.python.value
	pBedGetfasta.script = "file:scripts/bedtools/pBedGetfasta.py"
	return pBedGetfasta


@procfactory
def _pBedClosest():
	"""
	@name:
		pBedClosest
	@description:
		Similar to intersect, closest searches for overlapping features in A and B. In the event that no feature in B overlaps the current feature in A, closest will report the nearest (that is, least genomic distance from the start or end of A) feature in B. For example, one might want to find which is the closest gene to a significant GWAS polymorphism. Note that closest will report an overlapping feature as the closest that is, it does not restrict to closest non-overlapping feature. The following iconic cheatsheet summarizes the funcitonality available through the various optyions provided by the closest tool.
	@input:
		`afile:file`: The -a file
		`bfile:file`: The -b file
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`params`:   Other parameters for `bedtools closest`, default: ""
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedClosest               = Proc(desc = 'Find the closest elements')
	pBedClosest.input         = "afile:file, bfile:file"
	pBedClosest.output        = "outfile:file:{{i.afile | fn}}.closest.bt"
	pBedClosest.args.bedtools = params.bedtools.value
	pBedClosest.args.params   = Box()
	pBedClosest.script        = "file:scripts/bedtools/pBedClosest.py"
	return pBedClosest

@procfactory
def _pBedClosest2():
	"""
	@name:
		pBedClosest2
	@description:
		Multiple b-file version of pBedClosest
	@input:
		`afile:file`:   The -a file
		`bfiles:files`: The -b files
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`params`:   Other parameters for `bedtools closest`, default: ""
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedClosest2               = Proc(desc = 'Find the closest elements')
	pBedClosest2.input         = "afile:file, bfiles:files"
	pBedClosest2.output        = "outfile:file:{{i.afile | fn}}.closest.bt"
	pBedClosest2.args.bedtools = params.bedtools.value
	pBedClosest2.args.params   = Box()
	pBedClosest2.script        = "file:scripts/bedtools/pBedClosest2.py"
	return pBedClosest2

@procfactory
def _pBedFlank():
	"""
	@name:
		pBedFlank
	@description:
		`bedtools flank` will create two new flanking intervals for each interval in a BED file. Note that flank will restrict the created flanking intervals to the size of the chromosome (i.e. no start < 0 and no end > chromosome size).
	@input:
		`infile:file`:  The input file
		`gfile:file`:   The genome size file
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`params`:   Other parameters for `bedtools flank`, default: ""
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedFlank               = Proc(desc = 'Create two new flanking intervals for each interval in a BED file.')
	pBedFlank.input         = "infile:file"
	pBedFlank.output        = "outfile:file:{{i.infile | fn}}.flank.bed"
	pBedFlank.args.extend   = False
	pBedFlank.args.gsize    = params.gsize.value
	pBedFlank.args.params   = Box()
	pBedFlank.args.bedtools = params.bedtools.value
	pBedFlank.lang          = params.python.value
	pBedFlank.script        = "file:scripts/bedtools/pBedFlank.py"
	return pBedFlank

@procfactory
def _pBedIntersect():
	"""
	@name:
		pBedIntersect
	@description:
		By far, the most common question asked of two sets of genomic features is whether or not any of the features in the two sets overlap with one another. This is known as feature intersection. bedtools intersect allows one to screen for overlaps between two sets of genomic features. Moreover, it allows one to have fine control as to how the intersections are reported. bedtools intersect works with both BED/GFF/VCF and BAM files as input.
	@input:
		`afile:file` : The a file
		`bfile:file`: The b file
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`params`:   Other parameters for `bedtools intersect`, default: ""
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedIntersect                     = Proc(desc = 'The wrapper of "bedtools intersect" with single input b file.')
	pBedIntersect.input               = "afile:file, bfile:file"
	pBedIntersect.output              = "outfile:file:{{i.afile | fn}}.intersect.bt"
	pBedIntersect.args.bedtools       = params.bedtools.value
	pBedIntersect.args.params         = Box()
	pBedIntersect.lang                = params.python.value
	pBedIntersect.script              = 'file:scripts/bedtools/pBedIntersect.py'
	return pBedIntersect

@procfactory
def _pBedIntersect2():
	"""
	@name:
		pBedIntersect2
	@description:
		Multiple b-file version of pBedIntersect
	@input:
		`afile:file` : The a file
		`bfiles:files`: The b files
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`params`:   Other parameters for `bedtools intersect`, default: ""
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedIntersect2               = Proc(desc = 'The wrapper of "bedtools intersect" with multiple input b files.')
	pBedIntersect2.input         = "afile:file, bfiles:files"
	pBedIntersect2.output        = "outfile:file:{{i.afile | fn}}.intersect2.bt"
	pBedIntersect2.args.bedtools = params.bedtools.value
	pBedIntersect2.args.params   = Box()
	pBedIntersect2.lang          = params.python.value
	pBedIntersect2.script        = 'file:scripts/bedtools/pBedIntersect2.py'
	return pBedIntersect2

@procfactory
def _pBedMakewindows():
	"""
	@name:
		pBedMakewindows
	@description:
		Makes adjacent or sliding windows across a genome or BED file.
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`intype`:   The format of input file, whether is a "bed" file or "genome" size file. Default: `bed`
		`params`:   Other parameters for `bedtools makewindows`, default: `Box()`
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedMakewindows               = Proc(desc = 'Makes adjacent or sliding windows across a genome or BED file.')
	pBedMakewindows.input         = "infile:file"
	pBedMakewindows.output        = "outfile:file:{{i.infile | fn}}.window.bed"
	pBedMakewindows.args.params   = Box()
	pBedMakewindows.args.bedtools = params.bedtools.value
	pBedMakewindows.args.intype   = 'bed'
	pBedMakewindows.script        = "file:scripts/bedtools/pBedMakewindows.py"
	return pBedMakewindows

# region pBedMerge2
@procfactory
def _pBedMerge():
	"""
	@name:
		pBedMerge
	@description:
		`bedtools merge` combines overlapping or book-ended features in an interval file into a single feature which spans all of the combined features.
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable,               default: `<params.bedtools>`
		`params`  : Other parameters for `bedtools merge`, default: {}
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedMerge               = Proc(desc = 'Merge regions in a bed file using `bedtools merge`.')
	pBedMerge.input         = "infile:file"
	pBedMerge.output        = "outfile:file:{{i.infile | fn}}.merged.bed"
	pBedMerge.args.bedtools = params.bedtools.value
	pBedMerge.args.params   = Box()
	pBedMerge.lang          = params.python.value
	pBedMerge.script        = "file:scripts/bedtools/pBedMerge.py"
	# endregion
	return pBedMerge

# region pBedMerge2
@procfactory
def _pBedMerge2():
	"""
	@name:
		pBedMerge2
	@description:
		A multi-input file model of pBedMerge: Merge multiple input files.
	@input:
		`infiles:files`: The input files
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable,               default: `<params.bedtools>`
		`params`  : Other parameters for `bedtools merge`, default: {}
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedMerge2               = Proc(desc = 'A multi-input file model of `pBedMerge`: Merge multiple input files.')
	pBedMerge2.input         = "infiles:files"
	pBedMerge2.output        = "outfile:file:{{i.infiles | fs2name}}.merged.bed"
	pBedMerge2.args.bedtools = params.bedtools.value
	pBedMerge2.args.params   = Box()
	pBedMerge2.envs.fs2name  = fs2name
	pBedMerge2.lang          = params.python.value
	pBedMerge2.script        = "file:scripts/bedtools/pBedMerge2.py"
	# endregion
	return pBedMerge2

@procfactory
def _pBedMultiinter():
	"""
	@name:
		pBedMultiinter
	@description:
		Identifies common intervals among multiple BED/GFF/VCF files.
	@input:
		`infiles:files`: The input files
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`params`:   Other parameters for `bedtools multiinter`, default: ""
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedMultiinter               = Proc(desc = 'Identifies common intervals among multiple BED/GFF/VCF files.')
	pBedMultiinter.input         = "infiles:files"
	pBedMultiinter.output        = "outfile:file:{{i.infiles | fs2name}}.multiinter.bt"
	pBedMultiinter.args.bedtools = params.bedtools.value
	pBedMultiinter.args.params   = Box()
	pBedMultiinter.envs.fs2name  = fs2name
	pBedMultiinter.script        = "file:scripts/bedtools/pBedMultiinter.py"
	return pBedMultiinter

# region pBedRandom
@procfactory
def _pBedRandom():
	"""
	@name:
		pBedRandom
	@description:
		`bedtools random` will generate a random set of intervals in BED6 format. One can specify both the number (-n) and the size (-l) of the intervals that should be generated.
	@input:
		`gfile:file`: The genome size file
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable,    default: `<params.bedtools>`
		`seed`    : The seed for randomization, default: None
		`gsize`   : The chromsize file.
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedRandom               = Proc(desc = 'Generate a random set of intervals in BED format.')
	pBedRandom.input         = "l, n"
	pBedRandom.output        = "outfile:file:random.L{{i.l}}.N{{i.n}}.S{{args.seed}}.bed"
	pBedRandom.args.seed     = None
	pBedRandom.args.bedtools = params.bedtools.value
	pBedRandom.args.gsize    = params.gsize.value
	pBedRandom.lang          = params.python.value
	pBedRandom.script        = "file:scripts/bedtools/pBedRandom.py"
	# endregion
	return pBedRandom

@procfactory
def _pBedShift():
	"""
	@name:
		pBedShift
	@description:
		`bedtools shift` will move each feature in a feature file by a user-defined number of bases. While something like this could be done with an awk '{OFS="\t" print $1,$2+<shift>,$3+<shift>}', bedtools shift will restrict the resizing to the size of the chromosome (i.e. no features before 0 or past the chromosome end).
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The result file
	@args:
		`gsize`   : The genome size file. Default: `<params.gsize>`
		`bedtools`: The bedtools executable. Default: `<params.bedtools>`
		`params`  : Other parameters for `bedtools shift`. Default: ``
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedShift               = Proc(desc = '`bedtools shift` will move each feature in a feature file by a user-defined number of bases.')
	pBedShift.input         = "infile:file"
	pBedShift.output        = "outfile:file:{{infile | fn}}.shifted.bed"
	pBedShift.args.bedtools = params.bedtools.value
	pBedShift.args.gsize    = params.gsize.value
	pBedShift.args.params   = Box()
	pBedShift.script        = "file:scripts/bedtools/pBedShift.py"
	return pBedShift

@procfactory
def _pBedShuffle():
	"""
	@name:
		pBedShuffle
	@description:
		`bedtools shuffle` will randomly permute the genomic locations of a feature file among a genome defined in a genome file. One can also provide an exclusions BED/GFF/VCF file that lists regions where you do not want the permuted features to be placed. For example, one might want to prevent features from being placed in known genome gaps. shuffle is useful as a null basis against which to test the significance of associations of one feature with another.
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`params`  : Other parameters for `bedtools shuffle`, default: ""
		`gsize`   : The chromsize file. Default: `params.gsize`
		`n`       : Only return top `n` records (act like sampling). Default: `0`
			- `0`/`None` will return all records
			- `n<1` will return proportion of the records
			- `n>=1` will return top `n` records
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedShuffle               = Proc(desc = 'Randomly permute the genomic locations of a BED file using `bedtools shuffle`')
	pBedShuffle.input         = "infile:file"
	pBedShuffle.output        = "outfile:file:{{i.infile | fn}}.shuffled.bed"
	pBedShuffle.args.bedtools = params.bedtools.value
	pBedShuffle.args.n        = 0
	pBedShuffle.args.gsize    = params.gsize.value
	pBedShuffle.args.seed     = False
	pBedShuffle.args.params   = Box()
	pBedShuffle.lang          = params.python.value
	pBedShuffle.script        = "file:scripts/bedtools/pBedShuffle.py"
	return pBedShuffle

@procfactory
def _pBedSubtract():
	"""
	@name:
		pBedSubtract
	@description:
		`bedtools subtract` searches for features in B that overlap A. If an overlapping feature is found in B, the overlapping portion is removed from A and the remaining portion of A is reported. If a feature in B overlaps all of a feature in A, the A feature will not be reported.
	@input:
		`afile:file`: The a file
		`bfile:file`: The b file
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`params`:   Other parameters for `bedtools subtract`, default: ""
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedSubtract               = Proc(desc = '`bedtools subtract` searches for features in B that overlap A.')
	pBedSubtract.input         = "afile:file, bfile:file"
	pBedSubtract.output        = "outfile:file:{{afile | fn}}.subtracted.bed"
	pBedSubtract.args.bedtools = params.bedtools.value
	pBedSubtract.args.params   = Box()
	pBedSubtract.script        = "file:scripts/bedtools/pBedSubtract.py"
	return pBedSubtract

@procfactory
def _pBedWindow():
	"""
	@name:
		pBedWindow
	@description:
		Similar to `bedtools intersect`, `window` searches for overlapping features in A and B. However, window adds a specified number (1000, by default) of base pairs upstream and downstream of each feature in A. In effect, this allows features in B that are near features in A to be detected.
	@input:
		`afile:file`: The a file
		`bfile:file`: The b file
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`params`:   Other parameters for `bedtools window`, default: ""
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedWindow               = Proc(desc = 'Similar to `bedtools intersect`, `window` searches for overlapping features in A and B.')
	pBedWindow.input         = "afile:file, bfile:file"
	pBedWindow.output        = "outfile:file:{{afile | fn}}.window.bed"
	pBedWindow.args.bedtools = params.bedtools.value
	pBedWindow.args.params   = Box()
	pBedWindow.script        = "file:scripts/bedtools/pBedWindow.py"
	return pBedWindow

@procfactory
def _pBedGenomecov():
	"""
	@name:
		pBedGenomecov
	@description:
		`bedtools genomecov` computes histograms (default), per-base reports (-d) and BEDGRAPH (-bg) summaries of feature coverage (e.g., aligned sequences) for a given genome.
		NOTE: only bam file input implemented here.
	@input:
		`infile:file`: The bam file
	@output:
		`outfile:file`: The result file
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`params`:   Other parameters for `bedtools genomecov`, default: `Box(bg = True)`
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedGenomecov               = Proc(desc = '`bedtools genomecov` computes histograms (default), per-base reports (-d) and BEDGRAPH (-bg) summaries of feature coverage (e.g., aligned sequences) for a given genome.')
	pBedGenomecov.input         = "infile:file"
	pBedGenomecov.output        = "outfile:file:{{infile | fn}}.genomecov.bt"
	pBedGenomecov.args.bedtools = params.bedtools.value
	pBedGenomecov.args.params   = Box(bg = True)
	pBedGenomecov.script        = "file:scripts/bedtools/pBedGenomecov.py"
	return pBedGenomecov

@procfactory
def _pBedCluster():
	"""
	@name:
		pBedCluster
	@description:
		Similar to merge, cluster report each set of overlapping or "book-ended" features in an interval file. In contrast to merge, cluster does not flatten the cluster of intervals into a new meta-interval; instead, it assigns an unique cluster ID to each record in each cluster. This is useful for having fine control over how sets of overlapping intervals in a single interval file are combined.
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file with cluster id for each record
	@args:
		`bedtools`: The bedtools executable, default: `<params.bedtools>`
		`params`:   Other parameters for `bedtools cluster`, default: `Box()`
	@requires:
		[bedtools](http://bedtools.readthedocs.io/en/latest/index.html)
	"""
	pBedCluster               = Proc(desc = 'Similar to merge, cluster report each set of overlapping or "book-ended" features in an interval file.')
	pBedCluster.input         = "infile:file"
	pBedCluster.output        = "outfile:file:{{infile | fn}}.clustered.bt"
	pBedCluster.args.bedtools = params.bedtools.value
	pBedCluster.args.params   = Box()
	pBedCluster.lang          = params.python.value
	pBedCluster.script        = "file:scripts/bedtools/pBedCluster.py"
	return pBedCluster




from os import path
from glob import glob
from pyppl import Box
from bioprocs.utils import shell

indir   = {{i.indir | quote}}
samfile = {{i.samfile | quote}}
outdir  = {{o.outdir | quote}}
plink   = {{args.plink | quote}}
nthread = {{args.nthread | repr}}
keep    = {{args.keep | repr}}
fam     = {{args.fam | repr}}
samname = {{args.samid | repr}}
params  = {{args.params | repr}}

plink = shell.Shell(dict(plink = plink), equal = ' ').plink

bedfile = glob(path.join(indir, '*.bed'))[0]
input   = bedfile[:-4]
output  = path.join(outdir, path.basename(input))

# prepare sample file
if samname == 'both':
	pass
else:
	samplefile = path.join(outdir, 'samples.txt')
	with open(samfile, 'r') as fin, open(samplefile, 'w') as fout:
		for line in fin:
			fout.write("\t".join(line.strip().split("\t")[:1] * 2) + "\n")
	samfile = samplefile

params.bfile   = input
params.out     = output
params.nthread = nthread
params['make-bed'] = True

key = 'keep' if keep and fam else 'keep-fam' if keep and not fam else 'remove' if not keep and not fam else 'remove-fam'
params[key] = samfile

plink(**params).run()



from shutil import move, rmtree
from os import makedirs, path, symlink, remove
from sys import stdout, stderr
from pyppl import Box
from bioprocs.utils import runcmd, mem2, cmdargs

infile    = {{ i.infile | quote }}
outfile   = {{ o.outfile | quote }}
tmpdir    = {{ args.tmpdir | quote }}
tmpdir    = path.join (tmpdir, "{{proc.id}}.{{i.infile | fn}}.{{job.index}}")
doRmdup   = {{ args.rmdup }}

if not path.exists (tmpdir):
	makedirs (tmpdir)

params = {{args.params}}
try:
{% case args.tool %}
	########### biobambam
	{% when 'biobambam'%}
	if doRmdup:
		params['rmdup'] = 1
		params['D']     = "/dev/null"
	params['I'] = infile
	params['O'] = outfile
	params['tmpfile'] = path.join(tmpdir, 'tmp.')
	params['markthreads'] = {{args.nthread}}

	cmd = '{{args.biobambam_bamsort}} %s' % cmdargs(params, dash='', equal = '=')
	runcmd (cmd)

	########### sambamba
	{% when 'sambamba' %}
	if doRmdup:
		params['r'] = True
	params['t'] = {{args.nthread}}
	params['tmpdir'] = tmpdir

	cmd = '{{args.sambamba}} markdup %s "%s" "%s"' % (cmdargs(params), infile, outfile)
	runcmd (cmd)

	########### samtools
	{% when 'samtools' %}
	cmd = '{{args.samtools}} rmdup "%s" "%s"' % (infile, outfile)
	runcmd (cmd)

	########### picard
	{% when 'picard' %}
	mem = mem2({{ args.mem | quote }}, 'java')

	params['-Djava.io.tmpdir'] = tmpdir
	if doRmdup:
		params['REMOVE_DUPLICATES'] = "true"
	params['TMP_DIR'] = tmpdir
	params['I'] = infile
	params['O'] = outfile
	params['M'] = "/dev/null"

	mfile = "/dev/null"
	cmd = '{{args.picard}} MarkDuplicates %s %s %s ' % (mem, cmdargs(params, dash='', equal='='), rmdup)
	runcmd (cmd)

	########### bamutil
	{% when 'bamutil' %}
	if doRmdup:
		params['rmDups'] = True
	params['in']  = infile
	params['out'] = outfile
	cmd = '{{args.bamutil}} dedup %s' % cmdargs(params, equal = ' ')
	runcmd (cmd)

{% endcase %}
except Exception as ex:
	stderr.write ("Job failed: %s" % str(ex))
	raise
finally:
	rmtree (tmpdir)



from pyppl import Box
from bioprocs.utils import runcmd, cmdargs

params = {{args.params}}

####### htseq
{% if args.tool == 'htseq' %}

{% if i.infile.endswith('.bam') %}
params['f'] = 'bam'
{% endif %}

cmd = '{{args.htseq}} %s "{{i.infile}}" "{{args.refgene}}" > "{{o.outfile}}"' % (cmdargs(params))
runcmd (cmd)

{% endif %}



import requests

data = {{i.data}}
r    = requests.post({{i.url | quote}}, data)

rc   = int(r.status_code)
if rc != 200:
	raise Exception('Failed to POST to {{i.url}}, status code is %s' % rc)

with open({{o.outfile | quote}}, 'w') as fout:
	fout.write(r.text)


import docx
from pyppl import Box
from docx.shared import Inches, Pt
from docx.enum.dml import MSO_COLOR_TYPE, MSO_THEME_COLOR
from docx.enum.text import WD_ALIGN_PARAGRAPH, WD_LINE_SPACING, WD_BREAK_TYPE, WD_COLOR, WD_TAB_ALIGNMENT, WD_TAB_LEADER, WD_UNDERLINE
from docx.enum.style import WD_STYLE, WD_STYLE_TYPE
from docx.enum.section import WD_ORIENT, WD_SECTION
from docx.enum.table import WD_TABLE_ALIGNMENT, WD_TABLE_DIRECTION
from docx.enum.shape import WD_INLINE_SHAPE_TYPE
from os import path
from runpy import run_path
from bioprocs.utils import log2pyppl

infile    = {{i.infile | quote}}
outfile   = {{o.outfile | quote}}
codefiles = {{i.codes | repr}}
error     = {{args.error | repr}}
bcode     = {{args.bcode | repr}}
if not isinstance(bcode, list):
	bcode = [bcode]
bcode = '\n'.join(bcode) + '\n'
acode     = {{args.acode | repr}}
if not isinstance(acode, list):
	acode = [acode]
acode = '\n'.join(acode) + '\n'
section   = {{args.section | repr}}
if section:
	section['type']   = section.get('type', 'NEW_PAGE')
	section['orient'] = section.get('orient', 'PORTRAIT')
	section['margin'] = section.get('margin', [36] * 4)
	if isinstance(section['margin'], int):
		section['margin'] = [section['margin']] * 4
	elif len(section['margin']) == 2:
		section['margin'].extend(section['margin'])
	elif len(section['margin']) == 3:
		section['margin'].append(section['margin'][1])

def doSection(doc, section = section, new = True):
	if not section:
		return
	sec = doc.add_section() if new else doc.sections[0]
	sec.type = getattr(WD_SECTION, section['type'])
	sec.orientation = getattr(WD_ORIENT, section['orient'])
	_new_width, _new_height = sec.page_height, sec.page_width
	sec.page_width, sec.page_height = _new_width, _new_height
	sec.top_margin, sec.right_margin, sec.bottom_margin, sec.left_margin = (Pt(x) for x in section['margin'])
	return sec

if infile and path.isfile(infile):
	doc = docx.Document(infile)
	# do section
	sec = doSection(doc, new = True)
	exec(bcode, globals())
else:
	doc = docx.Document()
	sec = doSection(doc, new = False)
	if infile:
		doc.add_heading(infile, 0)
	exec(bcode, globals())

for codefile in codefiles:
	log2pyppl('Doing: {}'.format(path.basename(codefile)), level = 'Info')
	try:
		_ = run_path(codefile, globals())
	except Exception as ex:
		if error == 'exit':
			raise
		else:
			log2pyppl('Failed to run: {}'.format(codefile), level = 'Error')
			for line in str(ex).splitlines():
				log2pyppl('\t' + line)

exec(acode, globals())
doc.save(outfile)






from pyppl import Box
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile1  = {{i.infile1 | quote}}
infile2  = {{i.infile2 | quote}}
outfile1 = {{o.outfile1 | quote}}
outfile2 = {{o.outfile2 | quote}}
inopts1  = {{args.inopts1 | repr}}
inopts2  = {{args.inopts2 | repr}}

if inopts1.headCallback:
	inopts1.headCallback = eval(inopts1.headCallback)
if inopts2.headCallback:
	inopts2.headCallback = eval(inopts2.headCallback)
inopts1.attach = False
inopts2.attach = False

rnames1 = inopts1.get('rnames', True)
rnames2 = inopts2.get('rnames', True)
if 'rnames' in inopts1:
	del inopts1['rnames']
if 'rnames' in inopts2:
	del inopts2['rnames']

indata1 = TsvReader(infile1, **inopts1)
indata2 = TsvReader(infile2, **inopts2)

cnames1 = indata1.meta if not rnames1 else indata1.meta[1:]
cnames2 = indata2.meta if not rnames2 else indata2.meta[1:]
paired  = list(set(cnames1) & set(cnames2))
cnames1 = cnames2 = paired

if rnames1:
	cnames1 = [indata1.meta[0]] + cnames1
if rnames2:
	cnames2 = [indata2.meta[0]] + cnames2

cindex1 = [indata1.meta.index(c) for c in cnames1]
cindex2 = [indata2.meta.index(c) for c in cnames2]

outdata1 = TsvWriter(outfile1)
outdata2 = TsvWriter(outfile2)
outdata1.meta = cnames1
outdata2.meta = cnames2
outdata1.writeHead()
outdata2.writeHead()

for r1 in indata1:
	outdata1.write(r1[i] for i in cindex1)
outdata1.close()

for r2 in indata2:
	outdata2.write(r2[i] for i in cindex2)
outdata2.close()



"""A set of processes for Tumor heterogeneity analysis"""
from modkit import Modkit
from pyppl import Box, Proc
from . import rimport, params, delefactory, procfactory
from .utils import fs2name
Modkit().delegate(delefactory())

@procfactory
def _pSciClone():
	"""
	@input:
		muts: Somatic mutations. Could be one of:
			- Single(paired)-sample VCF files, separared by comma
				- Must have `FORMAT/AF` and `FORMAT/AD`
			- Multi-sample VCF file
			- TCGA MAF file
				- Must have `t_alt_count` and `t_ref_count`
			- SciClone TSV file
				- Chr, Pos, RefCount, VarCount, VAF*100
		cnvs: Copy numbers. Could be one of:
			- Single-sample VCF files, separared by comma
			- Multi-sample VCF files, must have `INFO/END` and `FORMAT/CN`
			- SciClone TSV file
				- Chr, Start, End, CopyNumber
	@output:
		outdir: The output directory.
	@args:
		params (Box)  : Other parameters for original `sciClone` function. Default: `Box()`
		exfile (file)  : The regions to be excluded. In BED3 format
		mutctrl (int|NoneType): Index of the control sample in paired-sample mutation VCF files or multi-sample VCF file, 0-based.
			- `None` indicates no control sample.
			- For paired-sample VCF files, if `None` specified, second sample(1) will be used as control.
		cnctrl (int|NoneType): Index of the control sample in copy number VCF files or multi-sample VCF file, 0-based.
	"""
	return Box(
		desc   = "Clonality analysis using SciClone.",
		input  = "muts:var, cnvs:var",
		output = "outdir:dir:{{i.muts | bn | .split('.')[0]}}.sciclone",
		lang   = params.Rscript.value,
		args   = Box(
			params  = Box(),
			exfile  = "",
			mutctrl = None,
			cnctrl  = None
		)
	)

@procfactory
def _pPyClone():
	"""
	@input:
		muts: Somatic mutations. Could be one of:
			- Single(paired)-sample VCF files, separated by comma (i.e. a.vcf,b.vcf)
				- Must have `FORMAT/AF` and `FORMAT/AD`
			- Multi-sample VCF file (i.e. a.vcf)
			- TCGA MAF file (i.e. a.maf)
				- Must have `t_alt_count` and `t_ref_count` columns
			- PyClone TSV file (i.e. a.tsv or a.txt)
				- See https://github.com/aroth85/pyclone/blob/master/examples/mixing/tsv/SRR385941.tsv without `*_cn` columns
				- `mutation_id` should be either `<sample>:<gt>:<chr>:<pos>` or `<chr>:<pos>`
		cnvs: Copy numbers. Could be one of:
			- Single-sample VCF files, separated by comma
			- Multi-sample VCF file
				- Must have `INFO/END` and `FORMAT/CN`
			- PyClone TSV file
				- See https://github.com/aroth85/pyclone/blob/master/examples/mixing/tsv/SRR385941.tsv without `*_counts` and `variant_freq` columns
	@output:
		`outdir:dir`: The output directory.
	@args:
		pyclone (str): Path to `PyClone`.
		bcftools(str): Path to `bcftools`, used to get information from VCF file.
		bedtools(str): Path to `bedtools`, used to intersect mutations with CNVs.
		params  (Box): Other parameters for original `PyClone run_analysis_pipeline` function.
		nthread (int): Number of threads to use by openblas.
		mutctrl (int|NoneType): Index of the control sample in paired-sample mutation VCF files or multi-sample VCF file, 0-based.
			- `None` indicates no control sample.
			- For paired-sample VCF files, if `None` specified, second sample(1) will be used as control.
		cnctrl (int|NoneType): Index of the control sample in copy number VCF files or multi-sample VCF file, 0-based.
	"""
	return Box(
		desc   = "Clonality analysis using PyClone",
		input  = "muts:var, cnvs:var",
		output = "outdir:dir:{{i.muts | bn | .split('.')[0]}}.pyclone",
		lang   = params.python.value,
		args   = Box(
			pyclone  = params.pyclone.value,
			bcftools = params.bcftools.value,
			bedtools = params.bedtools.value,
			refgene  = params.refgene.value,
			nthread  = 1,
			params   = Box(),
			mutctrl  = None,
			cnctrl   = None,
		)
	)

@procfactory
def _pAllFIT():
	"""
	@description:
		All-FIT - Allele-Frequency-based Imputation of Tumor Purity infers specimen purity from tumor-only samples sequenced with deep sequencing. It is developed in Khiabanian Lab by Jui Wan Loh and Hossein Khiabanian.
		See: https://github.com/KhiabanianLab/All-FIT
	@input:
		infile: The input file, could be one of:
			- Single(paired)-sample VCF file with `FORMAT/AF` and `FORMAT/AD`
			- All-FIT format (see: https://github.com/KhiabanianLab/All-FIT/blob/master/test/input/sampleFile1.xls)
		cnfile: The copy number variation file, could be one of:
			- Omitted (then Ploidy in `infile` must be provided)
			- Single(paired)-sample VCF file with `INFO/END` and `FORMAT/CN`
			- Bed3 file with 4th column as the copy number.
	@output:
		outfile: The output file with purity
		outdir: The output directory with output file and figures.
	@args:
		allfit (str): Path to All-FIT.py
		bcftools (str): Path to `bcftools`, used to get information from VCF file.
		bedtools (str): Path to `bedtools`, used to intersect mutations with CNVs.
		params (Box): Other parameters for All-FIT.py
		mutctrl (int|NoneType): Index of the control sample in paired-sample mutation VCF file, 0-based.
			- `None` indicates no control sample.
			- For paired-sample VCF files, if `None` specified, second sample(1) will be used as control.
		cnctrl (int|NoneType): Index of the control sample in copy number VCF file, 0-based.
		nthread (int): Number of threads to use for openblas.
	"""
	return Box(
		desc = "Allele-Frequency-based Imputation of Tumor Purity Inference",
		input = "infile:file, cnfile:var",
		output = [
			"outfile:file:{{i.infile | stem | @append: '.allfit'}}/{{i.infile | stem | @append: '.purity.txt'}}",
			"outdir:dir:{{  i.infile | stem | @append: '.allfit'}}"
		],
		lang = params.python.value,
		args = Box(
			allfit   = params.allfit.value,
			bcftools = params.bcftools.value,
			bedtools = params.bedtools.value,
			params   = Box(t = 'somatic'),
			mutctrl  = None,
			cnctrl   = None,
			nthread  = 1
		)
	)

@procfactory
def _pTMBurden():
	"""
	@input:
		infile: The input MAF file
	@output:
		outfile: The tumor mutation burden file
	@args:
		type: The type of mutation burden.
			- `nonsyn`: Counting nonsynonymous mutations
	"""
	return Box(
		desc    = 'Calculation of tumor mutation burden.',
		input   = 'infile:file',
		output  = 'outfile:file:{{i.infile | stem}}.tmb.txt',
		args    = Box(type = 'nonsyn'),
		lang    = params.python.value,
		mutctrl = None,
		cnctrl  = None,
	)

@procfactory
def _pQuantumClone():
	"""
	@description:
		Clonality analysis using QuantumClone:
		https://academic.oup.com/bioinformatics/article/34/11/1808/4802225
	@input:
		`vfvcfs:files`: The input vcf files with mutations
	@output:
		`outdir:dir`: The output directory
	@args:
		`params`  : other parameters for `QuantumClone`'s `One_step_clustering`
		`vfsamcol`: The index of the target sample in mutation VCF file, 1-based. Default: `1`
		`varcount`: An R function string to define how to get the variant allele count. Default: `function(fmt) as.integer(unlist(strsplit(fmt$AD, ","))[2])`
			- If this function returns `NULL`, record will be skipped.
			- It can use the sample calls (`fmt`) and also the record info (`info`)
			- Both `function(fmt) ...` and `function(fmt, info) ...` can be used.
			- Don't include `info` if not necessary. This saves time.
			- This function can return the variant count directly, or
			- an R `list` like: `list(count = <var count>, depth = <depth>)`.
			- By default, the `depth` will be read from `fmt$DP`
		`nthread` : # threads to use. Default: `1`
	"""
	pQuantumClone               = Proc(desc = "Clonality analysis using QuantumClone")
	pQuantumClone.input         = 'vfvcfs:files'
	pQuantumClone.output        = "outdir:dir:{{i.vfvcfs | fs2name}}.qclone"
	pQuantumClone.envs.fs2name  = fs2name
	pQuantumClone.args.params   = Box()
	pQuantumClone.args.vfsamcol = 1 # 1-based
	pQuantumClone.args.varcount = 'function(fmt) as.integer(unlist(strsplit(fmt$AD, ","))[2])'
	pQuantumClone.args.nthread  = 1
	pQuantumClone.lang          = params.Rscript.value
	return pQuantumClone

@procfactory
def _pTheta():
	"""
	@description:
		Run THetA2 for tumor purity calculation
		Set lower MIN_FRAC if interval is not enough and NO_CLUSTERING if it raises
		"No valid Copy Number Profiles exist", but have to pay attention to the results.
		(see: https://groups.google.com/forum/#!topic/theta-users/igrEUol3sZo)

		THetA2 needs an interval file with the coverages of both tumor and normal and SNP files
		with counts of mutation allele and reference allele for both tumor and normal as well.
	@input:
		infile: A bed3 file, or a THetA2 interval file.
			- See: https://github.com/raphael-group/THetA/blob/master/example/Example.intervals
		tumbam: bam file for tumor or SNP file for tumor.
			- See: https://raw.githubusercontent.com/raphael-group/THetA/master/example/TUMOR_SNP.formatted.txt
		normbam: bam file for normal or SNP file for normal.
			- See: https://raw.githubusercontent.com/raphael-group/THetA/master/example/NORMAL_SNP.formatted.txt
	@output:
		outfile: The output file with purity.
		outdir: The output directory with output files and figures.
	@args:
		theta (str): Path to THetA2
		bam_readcount (str): Path to bam_readcount.
		bedtools (str): Path to bedtools, used to extract coverage of regions.
		samtools (str): Path to samtools, used to index bam file if possible.
		ref (file): The reference genome file.
		params (Box): Other parameters for THetA2.
		nthread (int): The number of threads to use.
		affysnps (str): The affymetrix Array snps, or other candidate snp list, in BED6-like format
			- The first 6 columns should be in BED6 format
			- The 7th column is reference allele, and 8th column is mutation allele.
	"""
	return Box(
		desc   = "Tumor purity calculation using THetA2",
		lang   = params.python.value,
		input  = 'infile:file, tumbam:file, normbam:file',
		output = [
			'outfile:file:{{i.infile | stem}}.theta/{{i.infile | stem}}.purity.txt',
			'outdir:dir:{{i.infile | stem}}.theta'
		],
		args = Box(
			theta         = params.theta2.value,
			bedtools      = params.bedtools.value,
			bam_readcount = params.bam_readcount.value,
			samtools      = params.samtools.value,
			params        = Box(),
			ref           = params.ref.value,
			nthread       = 1,
			affysnps      = params.affysnps.value,
		)
	)

@procfactory
def _pSuperFreq():
	pSuperFreq              = Proc(desc = "Subclonal analysis with superFreq")
	pSuperFreq.input        = "indir:dir, gfile:file"
	pSuperFreq.output       = "outdir:dir:{{i.indir | fn2}}-{{i.gfile | fn2}}.superfreq"
	pSuperFreq.args.nthread = 1
	pSuperFreq.args.baits   = '' # target regions
	pSuperFreq.args.ref     = params.ref.value
	pSuperFreq.args.resdir  = params.superfreq_res.value
	pSuperFreq.args.genome  = params.genome.value
	pSuperFreq.args.params  = Box(
		systematicVariance = .02, maxCov = 150, BQoffset = 33,
		mode = 'exome', splitRun = True
	)
	pSuperFreq.lang         = params.Rscript.value
	return pSuperFreq

@procfactory
def _pClonEvol():
	"""
	@input:
		mutfile: The mutation file or output directory from PyClone.
			- https://github.com/hdng/clonevol/issues/4#issuecomment-280997440
			- VAF, CCF should x100
		samfile: The sample information file.
		drivers: The driver genes.
			- One per line or
			- Mutsig file with top 10 genes
	@output:
		outdir: The output directory.
	@args:
		inopts: The input options to read the mutation file.
		params: The parameters for individual `ClonEvol` functions.
	"""
	return Box(
		desc   = "Inferring and visualizing clonal evolution in multi-sample cancer sequencing",
		input  = 'mutfile:file, samfile:file, drivers:var',
		output = 'outdir:dir:{{i.mutfile | stem}}.clonevol',
		lang   = params.Rscript.value,
		args   = Box(
			# only for clonevol input format
			inopts   = Box(rnames = False,  cnames = True),
			drivers  = [],
			refgene  = params.refgene.value,
			bedtools = params.bedtools.value,
			devpars  = Box(width = 2000, height = 2000, res = 300),
			params   = Box({
				'plot.variant.clusters': Box(
					# see https://rdrr.io/github/hdng/clonevol/man/plot.variant.clusters.html
				),
				'plot.cluster.flow': Box({
					# see https://rdrr.io/github/hdng/clonevol/man/plot.cluster.flow.html
				}),
				'infer.clonal.models': Box({
					# see https://rdrr.io/github/hdng/clonevol/man/infer.clonal.models.html
					"founding.cluster": 1,
					"cluster.center": "mean",
					"sum.p.cutoff": 0.05,
					"alpha": 0.05,
				}),
				'transfer.events.to.consensus.trees': Box({
					# see https://rdrr.io/github/hdng/clonevol/man/transfer.events.to.consensus.trees.html
					"event.col.name": "gene"
				}),
				'convert.consensus.tree.clone.to.branch': Box({
					# see https://rdrr.io/github/hdng/clonevol/man/convert.consensus.tree.clone.to.branch.html
					"branch.scale": "sqrt"
				}),
				'plot.clonal.models': Box({
					# see https://rdrr.io/github/hdng/clonevol/man/plot.clonal.models.html
					# "clone.shape"                     : 'bell',
					# "bell.event"                      : True,
					# "bell.event.label.color"          : 'blue',
					# "bell.event.label.angle"          : 60,
					# "clone.time.step.scale"           : 1,
					# "bell.curve.step"                 : 2,
					# "merged.tree.plot"                : True,
					# "tree.node.label.split.character" : None,
					# "tree.node.shape"                 : 'circle',
					# "tree.node.size"                  : 30,
					# "tree.node.text.size"             : 0.5,
					# "merged.tree.node.size.scale"     : 1.25,
					# "merged.tree.node.text.size.scale": 2.5,
					# "merged.tree.cell.frac.ci"        : False,
					# "mtcab.event.sep.char"            : ',',
					"mtcab.branch.text.size"          : .8,
					# "mtcab.branch.width"              : 0.75,
					# "mtcab.node.size"                 : 3,
					# "mtcab.node.label.size"           : 1,
					"mtcab.node.text.size"            : .8,
					# "cell.plot"                       : True,
					# "num.cells"                       : 100,
					# "cell.border.size"                : 0.25,
					# "cell.border.color"               : 'black',
					# "clone.grouping"                  : 'horizontal',
					# "show.score"                      : False,
					# "cell.frac.ci"                    : True,
					# "disable.cell.frac"               : False
				})
			})
		)
	)

@procfactory
def _pSchism():
	"""
	@description:
		Infer subclonal hierarchy and the tumor evolution from somatic mutations using SCHISM.
		See: https://github.com/KarchinLab/SCHISM
	@input:
		infile: The input mutation file
			- In format of:
				```
				sampleID	mutationID	referenceReads	variantReads	copyNumber	clusterID
				S1	0	368	132	2	0
				S2	1	381	119	2	0
				S3	2	367	133	3	1
				...
				```
			- Or `pPyClone` output directory
			- Or a MAF file
		purity: The purity of the samples, two columns:
			- Sample[tab]Purity, without header.
			- If omitted, will assume 100% purity for all samples.
	@output:
		outdir: The output directory.
	@args:
		schism   (str) : Path to runSchism
		params   (Box) : Other parameters for runSchism in yaml config.
		fishplot (bool): Whether generate a fishplot or not, requires fishplot install with R.
			- Not implemented yet
		Rscript  (str) : Path to Rscript to run fishplot
	"""
	return Box(
		desc   = 'Infer subclonal hierarchy and the tumor evolution from somatic mutations using SCHISM',
		lang   = params.python.value,
		input  = 'infile:file, purity:file',
		output = 'outdir:dir:{{i.infile | stem}}.schism',
		args   = Box(
			schism   = params.schism.value,
			dot      = params.dot.value,
			fishplot = True,
			Rscript  = params.Rscript.value,
			devpars  = Box(res = 100),
			params   = Box(
				cellularity_estimator = Box(),
				hypothesis_test       = Box(),
				genetic_algorithm     = Box()
			)
		)
	)

@procfactory
def _pLichee():
	"""
	@description:
		Fast and scalable inference of multi-sample cancer lineages using LICHeE.
		See: https://github.com/pwwang/lichee
	@input:
		infile: The input file, could be one of:
			- LICHeE input file. Set `args.params.cp = True` if input data represents cell prevalence instead of VAF.
			- A pPyClone output directory.
			- A MAF file.
	@output:
		outdir: The output directory
	@args:
		dot (str): Path to dot to generate figures
		params (Box): Other parameters for `lichee`.
			- Set a larger `e` if you have noisy data. Default is 0.1.
	"""
	return Box(
		desc   = 'Fast and scalable inference of multi-sample cancer lineages using LICHeE',
		input  = 'infile:file',
		output = 'outdir:dir:{{i.infile | stem}}.lichee',
		lang   = params.python.value,
		args   = Box(
			dot      = params.dot.value,
			lichee   = params.lichee.value,
			fishplot = True,
			Rscript  = params.Rscript.value,
			devpars  = Box(res = 100),
			params   = Box(maxVAFAbsent = 0.005, minVAFPresent = 0.005)
		)
	)

@procfactory
def _pPyClone2ClonEvol():
	pPyClone2ClonEvol               = Proc(desc = "Convert PyClone results to ClonEvol input format.")
	pPyClone2ClonEvol.input         = 'indir:dir'
	pPyClone2ClonEvol.output        = 'outfile:file:{{i.indir | fn}}.clonevol.txt'
	pPyClone2ClonEvol.args.refgene  = params.refgene.value
	pPyClone2ClonEvol.args.drivers  = []
	pPyClone2ClonEvol.args.bedtools = params.bedtools.value
	pPyClone2ClonEvol.lang          = params.python.value
	return pPyClone2ClonEvol






"""
A set of processes to generate/process sam/bam files
"""
from pyppl import Proc, Box
#from .utils import mem2, runcmd, buildref, checkref, polling, helpers, plot, dirnameFiles
from . import params, bashimport, rimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pSam2Bam():
	"""
	@input:
		infile: The input file
	@output:
		outfile: The output bam file
	@args:
		tool             : The tool used to do the sort. Available: sambamba|picard|sambamba|biobambam|samtools
		sambamba         : The path of the sambamba.
		picard           : The path of the picard.
		biobambam_bamsort: The path of the biobambam's bamsort.
		samtools: The path of the samtools.
		sort    : Do sorting?
			- If input is sam, tool is biobambam, this should be True
		index  : Do indexing?
		markdup: Do duplicates marking?
			- `rmdup` for samtools will be called
		rmdup  : Do duplicates removing?
		tmpdir : The tmp dir used to store tmp files.
		sortby : Sort by coordinate or queryname.
		nthread: Number of threads to use.
		infmt  : The format of input file. Available: sam|bam
		params : Other parameters for `tool`.
		mem    : The max memory to use.
			- Unit could be G/g/M/m
			- Will be converted to -Xmx4G, and -Xms will be 1/8 of it
	@requires:
		[sambamba](https://lomereiter.github.io/sambamba/docs/sambamba-view.html) if `args.tool` == samtools or reference used but not indexed.
		[picard](https://broadinstitute.github.io/picard/command-line-overview.html)
		[biobambam](https://github.com/gt1/biobambam2)
		[samtools](https://github.com/samtools/samtools)
	"""
	return Box(
		desc   = 'Deal with mapped sam/bam files, including sort, markdup, rmdup, and/or index.',
		lang   = params.python.value,
		input  = "infile:file",
		output = "outfile:file:{{i.infile | fn}}.bam, outidx:file:{{i.infile | fn}}.bam.bai",
		errhow = 'retry',
		preCmd = """
			{{bashimport}} reference.bash
			export elprep={{args.elprep | quote}}
			if [[ {{args.tool | quote}} == "elprep" ]]; then
				reference elprep {{args.ref | squote}}
			fi""",
		args = Box(
			tool       = "elprep",
			sambamba   = params.sambamba.value,
			picard     = params.picard.value,
			biobambam  = params.biobambam_bamsort.value,
			samtools   = params.samtools.value,
			elprep     = params.elprep.value,
			steps      = Box(sort=True, index=True, markdup=True, rmdup=True, recal=True),
			tmpdir     = params.tmpdir.value,
			sortby     = "coordinate",
			nthread    = 1,
			params     = Box(),
			mem        = params.mem16G.value,
			ref        = params.ref.value,
			knownSites = ''))

@procfactory
def _pBamMarkdup():
	"""
	@name:
		pBamMarkdup
	@description:
		Mark/remove duplicates for bam files
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output bam file
	@args:
		`tool`             : The tool used to do the sort. Default: sambamba (picard|sambamba|biobambam|samtools|bamutil)
		`sambamba`         : The path of sambamba. Default: sambamba
		`picard`           : The path of picard. Default: picard
		`biobambam_bamsort`: The path of biobambam's bamsort. Default: bamsort
		`samtools`         : The path of samtools. Default: samtools
		`bamutil`          : The path of bamutil. Default: bam
		`rmdup`            : Do duplicates removing? Default: False
		- Samtools will anyway remove the duplicates
		`tmpdir`           : The tmp dir used to store tmp files. Default: <system default tmpdir>
		`nthread`          : Default: 1
		- Not available for samtools and picard
		`params`           : Other parameters for `tool`. Defaut: ""
		`mem`              : The max memory to use. Default: "16G"
		- Unit could be G/g/M/m
		- Will be converted to -Xmx4G, and -Xms will be 1/8 of it
	@requires:
		[sambamba](https://lomereiter.github.io/sambamba/docs/sambamba-view.html)
		[picard](https://broadinstitute.github.io/picard/command-line-overview.html)
		[biobambam](https://github.com/gt1/biobambam2)
		[samtools](https://github.com/samtools/samtools)
		[bamutil](http://genome.sph.umich.edu/wiki/BamUtil#Programs)
	"""
	pBamMarkdup                        = Proc(desc = 'Mark/remove duplicates for bam files.')
	pBamMarkdup.input                  = "infile:file"
	pBamMarkdup.output                 = "outfile:file:{{i.infile | fn}}.bam"
	pBamMarkdup.args.tool              = "biobambam"
	pBamMarkdup.args.sambamba          = params.sambamba.value
	pBamMarkdup.args.picard            = params.picard.value
	pBamMarkdup.args.biobambam_bamsort = params.biobambam_bamsort.value
	pBamMarkdup.args.samtools          = params.samtools.value
	pBamMarkdup.args.bamutil           = params.bamutil.value
	pBamMarkdup.args.rmdup             = False
	pBamMarkdup.args.tmpdir            = params.tmpdir.value
	pBamMarkdup.args.nthread           = 1
	pBamMarkdup.args.params            = Box()
	pBamMarkdup.args.mem               = params.mem16G.value
	#pBamMarkdup.envs.mem2              = mem2.py
	#pBamMarkdup.envs.runcmd            = runcmd.py
	#pBamMarkdup.envs.params2CmdArgs    = helpers.params2CmdArgs.py
	pBamMarkdup.lang                   = params.python.value
	pBamMarkdup.script                 = "file:scripts/sambam/pBamMarkdup.py"
	return pBamMarkdup

@procfactory
def _pBamRecal():
	"""
	@name:
		pBamRecal
	@description:
		Recalibrate a bam file
	@input:
		`infile:file`: The bam file
	@output:
		`outfile:file`: The output bam file
	@args:
		`tool`    : The tool used to recalibrate the bam file. Default: `gatk` (gatk|bamutil)
		`gatk`    : The path of gatk, including java path. Default: `gatk`
		`samtools`: The path of samtools. Default: `samtools`
		`bamutil` : The path of bamutil. Default: `bam`
		`picard`  : The path of picard. Default: `picard`
		`params`  : Other parameters for `bam recab`. Default         : ""
			`RealignerTargetCreator` : Other parameters for `gatk RealignerTargetCreator`. Defaut: ""
			`IndelRealigner`         : Other parameters for `gatk IndelRealigner`. Defaut: ""
			`BaseRecalibrator`       : Other parameters for `gatk BaseRecalibrator`. Defaut: ""
			`PrintReads`             : Other parameters for `gatk PrintReads`. Defaut: ""
		`mem`: The max memory to use. Default: "32G"
		`knownSites`: The known polymorphic sites to mask out. Default: "" (Required for GATK)
		`ref`: The reference file. Required.
			- Will be converted to -Xmx4G, and -Xms will be 1/8 of it
	@requires:
		[gatk](https://software.broadinstitute.org/gatk)
		[samtools](https://github.com/samtools/samtools) if `args.ref` is not indexed, or bamutil is used for bam index file generation.
		[picard](https://broadinstitute.github.io/picard/command-line-overview.html) if `args.ref is not dicted.`
	"""
	pBamRecal               = Proc(desc = 'Recalibrate a bam file.')
	pBamRecal.input         = "infile:file"
	pBamRecal.output        = "outfile:file:{{i.infile | fn}}.bam, outidx:file:{{i.infile | fn}}.bam.bai"
	pBamRecal.args.tool     = "bamutil"
	pBamRecal.args.gatk     = params.gatk.value
	pBamRecal.args.samtools = params.samtools.value
	pBamRecal.args.picard   = params.picard.value
	pBamRecal.args.bamutil  = params.bamutil.value
	pBamRecal.args.params   = Box(
		# For GATK
		# RealignerTargetCreator = Box(),
		# IndelRealigner         = Box(),
		# BaseRecalibrator       = Box(),
		# PrintReads             = Box()
	)
	pBamRecal.args.ref        = params.ref.value
	pBamRecal.args.tmpdir     = params.tmpdir.value
	pBamRecal.args.knownSites = params.dbsnp.value
	pBamRecal.args.nthread    = 1
	pBamRecal.args.mem        = params.mem32G.value
	pBamRecal.envs.bashimport = bashimport
	pBamRecal.preCmd          = """
	{{bashimport}} reference.bash
	export samtools={{args.samtools | squote}}
	reference fasta {{args.ref | squote}}
	"""
	pBamRecal.lang   = params.python.value
	pBamRecal.script = "file:scripts/sambam/pBamRecal.py"
	return pBamRecal

@procfactory
def _pBamReadGroup():
	"""
	@name:
		pBamReadGroup
	@description:
		Add or replace read groups of a bam file
	@input:
		`infile:file`: The bam file
	@output:
		`outfile:file`: The output bam file
	@args:
		`tool`                         : The tool used. Default: `picard` (picard|bamutil)
		`picard`                       : The path of picard. Default: `picard`
		`bamutil`                      : The path of bamutil. Default: `bam`
		`rg`                           : The read group. Default: {'id': '', 'pl': 'Illumina', 'pu': 'unit1', 'lb': 'lib1', 'sm': ''}
		- `id` will be parsed from filename with "_LX_" in it if not given
		- `sm` will be parsed from filename
		`params`                       : Other parameters for `tool`. Defaut: ""
		`mem`                          : The max memory to use. Default: "4G"
		- Will be converted to -Xmx4G, and -Xms will be 1/8 of it
		`tmpdir`                       : The temporary directory. Default: <system tmpdir>
	@requires:
		[gatk](https://lomereiter.github.io/sambamba/docs/sambamba-view.html)
		[samtools](https://github.com/samtools/samtools) if `args.ref` is not indexed.
		[picard](https://broadinstitute.github.io/picard/command-line-overview.html) if `args.ref is not dicted.`
	"""
	pBamReadGroup                     = Proc(desc = 'Add or replace read groups of a bam file.')
	pBamReadGroup.input               = "infile:file"
	pBamReadGroup.output              = "outfile:file:{{i.infile | bn}}"
	pBamReadGroup.args.tool           = "bamutil"
	pBamReadGroup.args.picard         = params.picard.value
	pBamReadGroup.args.bamutil        = params.bamutil.value
	pBamReadGroup.args.rg             = Box({'id': '', 'pl': 'Illumina', 'pu': 'unit1', 'lb': 'lib1', 'sm': ''})
	pBamReadGroup.args.params         = Box()
	pBamReadGroup.args.tmpdir         = params.tmpdir.value
	pBamReadGroup.args.mem            = params.mem4G.value
	#pBamReadGroup.envs.params2CmdArgs = helpers.params2CmdArgs.py
	pBamReadGroup.lang                = params.python.value
	pBamReadGroup.script              = "file:scripts/sambam/pBamReadGroup.py"
	return pBamReadGroup


@procfactory
def _pBamReorder():
	"""
	@name:
		pBamReorder
	@description:
		Reorder a sam/bam file by a given reference file using `picard ReorderSam`
	@input:
		`infile:file`: The sam/bam file
	@output:
		`outfile:file`: The output bam file
	@args:
		`picard`                       : The path of picard. Default: `picard`
		`ref`                          : The reference file. Required
		`params`                       : Other parameters for `picard ReorderSam`. Defaut: ""
		`mem`                          : The max memory to use. Default: "4G"
		- Will be converted to -Xmx4G, and -Xms will be 1/8 of it
		`tmpdir`                       : The temporary directory. Default: <system tmpdir>
	@requires:
		[picard](https://broadinstitute.github.io/picard/command-line-overview.html)
	"""
	pBamReorder                     = Proc(desc = 'Reorder a sam/bam file by a given reference.')
	pBamReorder.input               = "infile:file"
	pBamReorder.output              = "outfile:file:{{i.infile | bn}}"
	pBamReorder.args.picard         = params.picard.value
	pBamReorder.args.params         = Box()
	pBamReorder.args.tmpdir         = params.tmpdir.value
	pBamReorder.args.mem            = params.mem4G.value
	pBamReorder.args.ref            = params.ref.value
	pBamReorder.lang                = params.python.value
	pBamReorder.script              = "file:scripts/sambam/pBamReorder.py"
	return pBamReorder


@procfactory
def _pBamMerge():
	"""
	@name:
		pBamMerge
	@description:
		Merges multiple SAM and/or BAM files (must be sorted by coordinate) into a single file.
	@input:
		`infiles:files`: Input sam/bam files to be merged
	@output:
		`outfile:file`: The merged bam file
	@args:
		`tool`     : The tool used to merge. Default: bamutil (picard|samtools|sambamba)
		`picard`   : The path of picard. Default: `picard`
		`bamutil`  : The path of bamutil. Default: `bam`
		`samtools` : The path of samtools. Default: `samtools`
		`sambamba` : The path of sambamba. Default: `sambamba`
		`params`   : Other parameters for `tool`. Defaut: ""
		`mem`      : The max memory to use. Default: "4G"
		- Will be converted to -Xmx4G, and -Xms will be 1/8 of it, just for picard
		`tmpdir`   : The temporary directory. Default: <system tmpdir>
		`nthread`  : # threads to use. Default: 1
		- For picard, if nthread>1, USE_THREADING=true, otherwise USE_THREADING=false
	@requires:
		[picard](https://broadinstitute.github.io/picard/command-line-overview.html)
	"""
	pBamMerge                     = Proc(desc = 'Merges multiple SAM and/or BAM sorted files into a single file.')
	pBamMerge.input               = "infiles:files"
	pBamMerge.output              = "outfile:file:{{i.infiles[0] | fn2}}.etc.bam"
	pBamMerge.args.tool           = "picard"
	pBamMerge.args.picard         = params.picard.value
	pBamMerge.args.bamutil        = params.bamutil.value
	pBamMerge.args.samtools       = params.samtools.value
	pBamMerge.args.sambamba       = params.sambamba.value
	pBamMerge.args.params         = Box()
	pBamMerge.args.tmpdir         = params.tmpdir.value
	pBamMerge.args.nthread        = 1
	pBamMerge.args.mem            = params.mem4G.value
	#pBamMerge.envs.fsDirname      = dirnameFiles
	pBamMerge.lang                = params.python.value
	pBamMerge.script              = "file:scripts/sambam/pBamMerge.py"
	return pBamMerge


@procfactory
def _pBam2Gmut():
	"""
	@name:
		pBam2Gmut
	@description:
		Call germline (snps and indels) from a call-ready bam file.
	@input:
		`infile:file`: The input bam file
	@output:
		`outfile:file`: The vcf file containing the mutations
	@args:
		`tool`      : The tool used to call mutations. Default: gatk (vardict, snvsniffer, platypus, strelka)
		`gatk`      : The path of gatk. Default: gatk
		`vardict`   : The path of vardict. Default: vardict
		`snvsniffer`: The path of snvsniffer. Default: SNVSniffer
		`samtools`  : The path of samtools. Default: samtools (used to generate reference index)
		`platypus`  : The path of platypus. Default: platypus
		`strelka`   : The path of strelka. Default: configureStrelkaGermlineWorkflow.py
		`cfgParams` : The params for `strelka` configuration. Default: ""
		`picard`    : The path of picard. Default: picard
		`mem`       : The memory to be used. Default: 32G
			- will be converted to -Xms4G -Xmx32G for java programs
		`ref`:          The reference file. Required.
		`gz`:           Gzip output file? Default: False
		`tmpdir`:       The temporary directory. Default: <system tmpdir>
		`params`:       Other params for `tool`. Default: ""
	@requires:
		[gatk](https://lomereiter.github.io/sambamba/docs/sambamba-view.html)
		[samtools](https://github.com/samtools/samtools) if `args.ref` is not indexed.
		[picard](https://broadinstitute.github.io/picard/command-line-overview.html) if `args.ref is not dicted.`
		[vardict](https://github.com/AstraZeneca-NGS/VarDict)
		[snvsniffer](http://snvsniffer.sourceforge.net/homepage.htm#latest)
		[platypus](http://www.well.ox.ac.uk/platypus)
		[strelka@2.7.1+](https://github.com/Illumina/strelka)
	"""
	pBam2Gmut                 = Proc(desc = 'Call germline (snps and indels) from a call-ready bam file.')
	pBam2Gmut.input           = "infile:file"
	pBam2Gmut.output          = "outfile:file:{{i.infile | fn}}.vcf{{args.gz | lambda x: '.gz' if x else ''}}"
	pBam2Gmut.args.tool       = "strelka"
	pBam2Gmut.args.gatk       = params.gatk.value
	pBam2Gmut.args.vardict    = params.vardict.value
	pBam2Gmut.args.snvsniffer = params.snvsniffer.value
	pBam2Gmut.args.samtools   = params.samtools.value # required by SNVSniffer to generate a bam header file
	pBam2Gmut.args.platypus   = params.platypus.value
	pBam2Gmut.args.picard     = params.picard.value
	pBam2Gmut.args.strelka    = params.strelka_germ.value
	pBam2Gmut.args.mem        = params.mem24G.value
	pBam2Gmut.args.ref        = params.ref.value
	pBam2Gmut.args.tmpdir     = params.tmpdir.value
	pBam2Gmut.args.cfgParams  = Box() # only for strelka
	pBam2Gmut.args.params     = Box()
	pBam2Gmut.args.gz         = False
	pBam2Gmut.args.nthread    = 1 # for gatk and platypus
	pBam2Gmut.envs.bashimport = bashimport
	pBam2Gmut.preCmd          = """
	{{bashimport}} reference.bash
	export samtools={{args.samtools | squote}}
	export picard={{args.picard | squote}}
	reference fasta {{args.ref | squote}}
	reference picard {{args.ref | squote}}
	"""
	pBam2Gmut.lang   = params.python.value
	pBam2Gmut.script = "file:scripts/sambam/pBam2Gmut.py"
	return pBam2Gmut

@procfactory
def _pBamPair2Smut():
	"""
	@name:
		pBamPair2Smut
	@description:
		Call somatic mutations from tumor-normal bam pair.
	@input:
		`tumor:file`: The tumor bam file
		`normal:file`: The normal bam file
	@output:
		`outfile:file`: The vcf file
	@args:
		`tool`: The tool used to call mutations. Default: gatk (somaticsniper, strelka, snvsniffer, virmid, varidct)
		`gatk`: The path to gatk. Default: gatk
		`somaticsniper`: The path to gatk. Default: bam-somaticsniper
		`strelka`: The path to gatk. Default: configureStrelkaSomaticWorkflow.py
		`snvsniffer`: The path to gatk. Default: SNVSniffer
		`virmid`: The path to gatk. Default: virmid
		`vardict`: The path to gatk. Default: vardict
		`samtools`: The path to gatk. Default: samtools
		`picard`: The path to gatk. Default: picard
		`configParams`: The configuration parameters for `configureStrelkaSomaticWorkflow.py`. Default: `{}`
		`params`: The parameters for main programs. Default: `{}`
		`meme`: The memory. Default: 24G
		`ref`: The reference genom. Default: `params.ref.value`
		`gz`: Whether gzip the output vcf file. Default: False
		`nthread`: The number of threads to use. Default: 1
		`tmpdir`: The temporary directory. Default: `params.tmpdir.value`
	@requires:
		[gatk](https://lomereiter.github.io/sambamba/docs/sambamba-view.html)
		[samtools](https://github.com/samtools/samtools) if `args.ref` is not indexed.
		[picard](https://broadinstitute.github.io/picard/command-line-overview.html) if `args.ref is not dicted.`
		[vardict](https://github.com/AstraZeneca-NGS/VarDict)
		[snvsniffer](http://snvsniffer.sourceforge.net/homepage.htm#latest)
		[platypus](http://www.well.ox.ac.uk/platypus)
		[strelka@2.7.1+](https://github.com/Illumina/strelka)
	"""
	pBamPair2Smut                    = Proc(desc = 'Call somatic mutations from tumor-normal bam pair.')
	pBamPair2Smut.input              = "tumor:file, normal:file"
	pBamPair2Smut.output             = "outfile:file:{{i.tumor | fn}}-{{i.normal | fn}}.vcf{{args.gz | lambda x: '.gz' if x else ''}}"
	pBamPair2Smut.args.tool          = 'strelka'
	pBamPair2Smut.args.gatk          = params.gatk.value # required for strelka
	pBamPair2Smut.args.somaticsniper = params.somaticsniper.value
	pBamPair2Smut.args.strelka       = params.strelka_soma.value # @2.7.1
	pBamPair2Smut.args.snvsniffer    = params.snvsniffer.value
	pBamPair2Smut.args.virmid        = params.virmid.value
	pBamPair2Smut.args.vardict       = params.vardict.value
	pBamPair2Smut.args.samtools      = params.samtools.value
	pBamPair2Smut.args.picard        = params.picard.value
	pBamPair2Smut.args.configParams  = Box() # only for strelka
	pBamPair2Smut.args.params        = Box()
	pBamPair2Smut.args.mem           = params.mem24G.value
	pBamPair2Smut.args.ref           = params.ref.value
	pBamPair2Smut.args.gz            = False
	pBamPair2Smut.args.nthread       = 1
	pBamPair2Smut.args.tmpdir        = params.tmpdir.value
	pBamPair2Smut.envs.bashimport    = bashimport
	pBamPair2Smut.preCmd             = """
	{{bashimport}} reference.bash
	export samtools={{args.samtools | squote}}
	export picard={{args.picard | squote}}
	reference fasta {{args.ref | squote}}
	reference picard {{args.ref | squote}}
	"""
	pBamPair2Smut.lang   = params.python.value
	pBamPair2Smut.script = "file:scripts/sambam/pBamPair2Smut.py"
	return pBamPair2Smut

@procfactory
def _pBamStats():
	"""
	@name:
		pBamStats
	@description:
		Get read depth from bam files.
	@input:
		`infile:file`: The input bam file
	@output:
		`outfile:file`: The output statistic file
		`outdir:dir`:   The directory containing result files and figures.
	@args:
		`tool`: The tool used to do the job. Default: bamstats
		`bamstats`: The path to bamstats. Default: bamstats
		`params`: Other params to main program. Default: `{}`
		`mem`: The memory to be used. Default: 16G
		`plot`: Whether plot the result. Default: True
	"""
	pBamStats                  = Proc(desc = 'Get read depth from bam files.')
	pBamStats.input            = 'infile:file'
	pBamStats.output           = 'outfile:file:{{i.infile | fn}}/{{i.infile | fn}}.stat.txt, outdir:dir:{{i.infile | fn}}'
	pBamStats.args.tool        = 'bamstats'
	pBamStats.args.bamstats    = params.bamstats.value
	pBamStats.args.params      = Box()
	pBamStats.args.mem         = params.mem16G.value
	pBamStats.args.plot        = True
	pBamStats.args.histplotggs = Box(xlab = {0: 'Read depth'}, ylab = {0: '# samples'})
	pBamStats.args.boxplotggs  = Box(xlab = {0: 'Counts'})
	pBamStats.args.devpars     = Box(res = 300, width = 2000, height = 2000)
	pBamStats.args.cap         = 500
	pBamStats.args.cutoff      = 0
	pBamStats.args.nfeats      = 40
	pBamStats.args.feature     = 'wgs'
	pBamStats.envs.rimport     = rimport
	pBamStats.preCmd        = """
	if [[ "{{args.plot | R}}" == "TRUE" && {{proc.forks}} -lt 2 ]]; then
		echo "Plots can only be done with proc.forks >= 2." 1>&2
		exit 1
	fi
	"""
	pBamStats.lang   = params.Rscript.value
	pBamStats.script = "file:scripts/sambam/pBamStats.r"
	return pBamStats

@procfactory
def _pBam2Fastq():
	"""
	@input:
		infile: The sam/bam file.
			- Sam files only available for biobambam, picard
	@output:
		fqfile1: The 1st match of paired reads
		fqfile2: The 2nd match of paired reads
	@args:
		`tool`     : The tool to use. Default: biobambam (bedtools, samtools, picard)
		`biobambam`: The path of bamtofastq of biobambam. Default: bamtofastq
		`bedtools` : The path of bedtools. Default: bedtools
		`samtools` : The path of samtools. Default: samtools
		`picard`   : The path of picard. Default: picard
		`mem`      : The memory to be used by picard. Default: 8G
		`gz`       : Whether gzip the output files. Default: True
		`params`:  : Other params for `tool`. Default: ''
		`tmpdir`   : The tmpdir. Default: `__import__('tempfile').gettempdir()`
	@requires:
		[picard](https://broadinstitute.github.io/picard/command-line-overview.html)
		[biobambam](https://github.com/gt1/biobambam2)
		[samtools](https://github.com/samtools/samtools)
		[bedtools](http://bedtools.readthedocs.io/en/latest/content/bedtools-suite.html)
	"""
	return Box(
		desc   = 'Convert sam/bam files to pair-end fastq files.',
		input  = "infile:file",
		output = [
			"fqfile1:file:{{ i.infile | fn }}_1.fastq{% if args.gz %}.gz{% endif %}",
			"fqfile2:file:{{ i.infile | fn }}_2.fastq{% if args.gz %}.gz{% endif %}"
		],
		lang = params.python.value,
		args = Box(
			tool      = 'biobambam',
			biobambam = params.biobambam_bamtofastq.value,
			bedtools  = params.bedtools.value,
			samtools  = params.samtools.value,
			picard    = params.picard.value,
			mem       = params.mem8G.value, # only for picard
			gz        = False,
			params    = Box(),
			tmpdir    = params.tmpdir.value))

@procfactory
def _pBam2FastqSE():
	"""
	@name:
		pBam2FastqSE
	@description:
		Convert sam/bam files to single-end fastq files.
	@input:
		`infile:file`: The sam/bam file.
			- Sam files only available for biobambam, picard
	@output:
		`fqfile:file`: The fastq file
	@args:
		`tool`     : The tool to use. Default: biobambam (bedtools, samtools, picard)
		`biobambam`: The path of bamtofastq of biobambam. Default: bamtofastq
		`bedtools` : The path of bedtools. Default: bedtools
		`samtools` : The path of samtools. Default: samtools
		`picard`   : The path of picard. Default: picard
		`mem`      : The memory to be used by picard. Default: 8G
		`gz`       : Whether gzip the output files. Default: True
		`params`:  : Other params for `tool`. Default: ''
		`tmpdir`   : The tmpdir. Default: `__import__('tempfile').gettempdir()`
	@requires:
		[picard](https://broadinstitute.github.io/picard/command-line-overview.html)
		[biobambam](https://github.com/gt1/biobambam2)
		[samtools](https://github.com/samtools/samtools)
		[bedtools](http://bedtools.readthedocs.io/en/latest/content/bedtools-suite.html)
	"""
	pBam2FastqSE                     = Proc(desc = 'Convert bam files to single-end fastq files.')
	pBam2FastqSE.input               = "infile:file"
	pBam2FastqSE.output              = "fqfile:file:{{i.infile | fn }}.fq{% if args.gz %}.gz{% endif %}"
	pBam2FastqSE.args.tool           = 'biobambam'
	pBam2FastqSE.args.biobambam      = params.biobambam_bamtofastq.value
	pBam2FastqSE.args.bedtools       = params.bedtools.value
	pBam2FastqSE.args.samtools       = params.samtools.value
	pBam2FastqSE.args.picard         = params.picard.value
	pBam2FastqSE.args.mem            = params.mem8G.value # only for picard
	pBam2FastqSE.args.gz             = False
	pBam2FastqSE.args.params         = Box()
	pBam2FastqSE.args.tmpdir         = params.tmpdir.value
	#pBam2FastqSE.envs.runcmd         = runcmd.py
	#pBam2FastqSE.envs.mem2           = mem2.py
	#pBam2FastqSE.envs.params2CmdArgs = helpers.params2CmdArgs.py
	pBam2FastqSE.lang                = params.python.value
	pBam2FastqSE.script              = "file:scripts/sambam/pBam2FastqSE.py"
	return pBam2FastqSE

@procfactory
def _pBam2Counts():
	"""
	@name:
		pBam2Counts
	@description:
		Extract read counts from RNA-seq bam files.
	@input:
		`infile:file`: The input bam files
	@outfile:
		`outfile:file`: The count file
	@args:
		`tool`: The tool used to extract counts. Default: ht-seq
		`htseq`: The path of htseq-count.
		`params`: Other params for main program.
		`refgene`: The reference gene in GTF format.
	@requires:
		[`htseq`](https://htseq.readthedocs.io/)
	"""
	pBam2Counts                     = Proc(desc = 'Extract read counts from RNA-seq bam files.')
	pBam2Counts.input               = 'infile:file'
	pBam2Counts.output              = 'outfile:file:{{i.infile | fn}}.counts'
	pBam2Counts.args.tool           = 'htseq'
	pBam2Counts.args.htseq          = params.htseq_count.value
	pBam2Counts.args.params         = Box()
	pBam2Counts.args.refgene        = params.refexon.value
	pBam2Counts.lang                = params.python.value
	pBam2Counts.script              = "file:scripts/sambam/pBam2Counts.py"
	return pBam2Counts

@procfactory
def _pBamIndex():
	"""
	@name:
		pBamIndex
	@description:
		Index bam files.
	@input:
		`infile:file`: The input bam file
	@output:
		`outfile:file`: The symbolic link to the input file
		`outidx:file` : The index file
	@args:
		`samtools`: Path to samtools. Default: `params.samtools`
		`params`  : Other parameters for samtools. Default: `Box(b = True)`
		`nthread` : # threads to use. Default: `1`
	"""
	pBamIndex               = Proc(desc = 'Index bam files')
	pBamIndex.input         = 'infile:file'
	pBamIndex.output        = 'outfile:file:{{i.infile | bn}}, outidx:file:{{i.infile | bn}}.bai'
	pBamIndex.args.tool     = 'samtools'
	pBamIndex.args.samtools = params.samtools.value
	pBamIndex.args.params   = Box(b = True)
	pBamIndex.args.nthread  = 1
	pBamIndex.lang          = params.python.value
	pBamIndex.script        = "file:scripts/sambam/pBamIndex.py"
	return pBamIndex




"""Processes for WORD files"""
from pyppl import Proc, Box
from . import params, rimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pDocx():
	"""
	@name:
		pDocx
	@description:
		Operating .docx file
	@input:
		`infile`: A input docx file or a string that used as Heading 0 of a new file
		`codes:files`: A set of files containing python codes operating the docs file.
			- Variable `doc` is used to pass across files.
	@output:
		`outfile:file`: The output docx file
	@args:
		`bcode`: Some extra BEFORE the content is inserted (after heading).
		`acode`: Some extra AFTER the content is inserted.
		`error`: What to do when error happens. Default: `exit`
			- `ignore` to add nothing to the document
		`section`: Start a new section of this table. Could be a box of values. Default: `Box()`.
			- `Box()` means no new section will start.
			- `type`: The type of the section. Could be one of:
				- `CONTINUOUS`
				- `NEW_COLUMN`
				- `NEW_PAGE` (default if not provided when `args.section.type` is not `None`)
				- `EVEN_PAGE` and
				- `ODD_PAGE`
			- `orient`: The orientation of the section. Default: `PORTRAIT`
				- Could also be `LANDSCAPE`
			- `margin`: The margin of the section, in points. Could be:
				- Single value for all margins. Default: 36
				- Paired values for top/bottom and left/right margins
				- 3 values for top, left/right and bottom margins.
				- 4 values for top, right, bottom and left margins
	@requires:
		[`python-docx`](http://python-docx.readthedocs.io/en/latest)
	"""
	pDocx              = Proc(desc = 'Operating .docx file')
	pDocx.input        = 'infile, codes:files'
	pDocx.output       = 'outfile:file:{{bn(i.infile) if path.isfile(i.infile) else str2fn(i.infile) + ".docx"}}'
	pDocx.args.bcode   = []
	pDocx.args.acode   = []
	pDocx.args.error   = 'exit' # or ignore
	pDocx.args.section = Box()
	pDocx.envs.str2fn  = lambda s, re = __import__('re'): re.sub(r'[^\w\-_\.]', '_', s)[:32] if s else 'doc'
	pDocx.envs.path    = __import__("os").path
	pDocx.lang         = params.python.value
	pDocx.script       = "file:scripts/docx/pDocx.py"
	return pDocx

@procfactory
def _pTable2DocxCode():
	"""
	@name:
		pTable2DocxCode
	@description:
		Convert a table to docx code
	@input:
		`infile`: The table data file
	@output:
		`outfile:file`: The code file
	@args:
		`font`   : The font of table content. Default: `Box(family: 'Arial', size: 10)`
		`section`: Start a new section of this table. Could be a box of values. Default: `Box()`.
			- `Box()` means no new section will start.
			- `type`: The type of the section. Could be one of:
				- `CONTINUOUS`
				- `NEW_COLUMN`
				- `NEW_PAGE` (default if not provided when `args.section.type` is not `None`)
				- `EVEN_PAGE` and
				- `ODD_PAGE`
			- `orient`: The orientation of the section. Default: `PORTRAIT`
				- Could also be `LANDSCAPE`
			- `margin`: The margin of the section, in points. Could be:
				- Single value for all margins. Default: 36
				- Paired values for top/bottom and left/right margins
				- 3 values for top, left/right and bottom margins.
				- 4 values for top, right, bottom and left margins
		`style`  : The style of the table. Default: `Light List Accent 1`
			- `Table Normal`
			- `Colorful Grid`, `Colorful Grid Accent 1~6`
			- `Colorful List`, `Colorful List Accent 1~6`
			- `Colorful Shading`, `Colorful Shading Accent 1~6`
			- `Dark List`, `Dark List Accent 1~6`
			- `Light Grid`, `Light Grid Accent 1~6`
			- `Light List`, `Light List Accent 1~6`
			- `Light Shading`, `Light Shading Accent 1~6`
			- `Medium Grid 1~3`, `Medium Grid 1~3 Accent 1~6`
			- `Medium List 1~2`, `Medium List 1~2 Accent 1~6`
			- `Medium Shading 1~2`, `Medium Shading 1~2 Accent 1~6`
			- `Table Grid`
		`title`  : The title of the table, could be heading to table caption, Could be:
			- A string, defaulted as `Heading 1` (2nd level heading)
			- A `tuple` or `list` with first element (string) as the content, 2nd element as the heading level
		`caption`: The caption of the table. Default: `None`
		`bcode`  : Some extra BEFORE the table is inserted.
		`acode`  : Some extra AFTER the table is inserted.
		`inopts` : The options for reading the input file.
	@requires:
		[`python-docx` v0.8.7](http://python-docx.readthedocs.io/en/latest)
	"""
	pTable2DocxCode              = Proc(desc = 'Convert a table to docx code')
	pTable2DocxCode.input        = 'infile:file'
	pTable2DocxCode.output       = 'outfile:file:{{i.infile | fn2}}.docxcode.py'
	pTable2DocxCode.args.font    = Box(family = 'Arial', size = 10)
	pTable2DocxCode.args.section = Box()
	pTable2DocxCode.args.style   = 'Light List Accent 1'
	pTable2DocxCode.args.title   = None
	pTable2DocxCode.args.caption = None
	pTable2DocxCode.args.inopts  = Box(cnames = True, skip = 0, delimit = "\t")
	pTable2DocxCode.args.bcode   = []
	pTable2DocxCode.args.acode   = []
	pTable2DocxCode.envs.rimport = rimport
	pTable2DocxCode.lang         = params.python.value
	pTable2DocxCode.script       = "file:scripts/docx/pTable2DocxCode.py"
	return pTable2DocxCode

@procfactory
def _pImage2DocxCode():
	"""
	@name:
		pImage2DocxCode
	@description:
		Convert an image to docx code
	@input:
		`infile`: The image
	@output:
		`outfile:file`: The code file
	@args:
		`scale`: The scale of the image. Default: `Box()`
			- With keys `width` and/or `height` in points.
			- If only one is set, the other is automatically scaled.
		`title`  : The title of the table, could be heading to table caption, Could be:
			- A string, defaulted as `Heading 1` (2nd level heading)
			- A `tuple` or `list` with first element (string) as the content, 2nd element as the heading level
			- Template available. i.e. `("Title: {{i.infile | fn2}}", 2)`
		`section`: Start a new section of this table. Could be a box of values. Default: `Box()`.
			- `Box()` means no new section will start.
			- `type`: The type of the section. Could be one of:
				- `CONTINUOUS`
				- `NEW_COLUMN`
				- `NEW_PAGE` (default if not provided when `args.section.type` is not `None`)
				- `EVEN_PAGE` and
				- `ODD_PAGE`
			- `orient`: The orientation of the section. Default: `PORTRAIT`
				- Could also be `LANDSCAPE`
			- `margin`: The margin of the section, in points. Could be:
				- Single value for all margins. Default: 36
				- Paired values for top/bottom and left/right margins
				- 3 values for top, left/right and bottom margins.
				- 4 values for top, right, bottom and left margins
		`legend`: The legend of the image. Default: `None`
			- If starts with `Figure X. `, this part will be bolded.
			- Template available. i.e. `Figure 1. {{i.infile | fn2}} shows good results.`
		`align `: How the image is aligned if it is not an inline element. Default: `CENTER`
			- To set it inline, set `args.align` as `None`, and add something to the paragraph(`para`) in `args.acode`.
		`bcode` : Some extra BEFORE the table is inserted.
			- `doc` is available for the `document` object
			- `sec` is available for the `section` object if `args.section` has been set.
		`acode` : Some extra AFTER the table is inserted.
			- `doc` and `sec` are still available
			- `para`: The paragraph where the run of the image is placed
			- `run`: The run where the image is placed
	@requires:
		[`python-docx`](http://python-docx.readthedocs.io/en/latest)
	"""
	pImage2DocxCode              = Proc(desc = 'Convert an image to docx code')
	pImage2DocxCode.input        = 'infile:file'
	pImage2DocxCode.output       = 'outfile:file:{{i.infile | fn2}}.docxcode.py'
	pImage2DocxCode.args.scale   = Box() # width and/or height in inches
	pImage2DocxCode.args.section = Box()
	pImage2DocxCode.args.title   = None
	pImage2DocxCode.args.legend  = None
	pImage2DocxCode.args.align   = 'CENTER'
	pImage2DocxCode.args.bcode   = []
	pImage2DocxCode.args.acode   = []
	pImage2DocxCode.lang         = params.python.value
	pImage2DocxCode.script       = "file:scripts/docx/pImage2DocxCode.py"
	return pImage2DocxCode




"""Tabix utilities"""
from pyppl import Proc, Box
#from .utils import helpers, runcmd
from . import params
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pTabix():
	"""
	@name:
		pTabix
	@description:
		Use tabix to extract information.
	@input:
		`infile`: a local or remote file
		`region`: a region or a file containing regions
	@output:
		`outfile:file`: The information extracted from the input file
	@args:
		`tabix`: The path to `tabix`
		`params`: Other params for `tabix`
	"""
	pTabix             = Proc(desc = 'Use tabix to extract information.')
	pTabix.input       = "infile, region"
	pTabix.output      = "outfile:file:{{i.infile | fn | fn}}-{{job.index}}{% if i.infile.endswith('.gz') %}{{i.infile | fn | ext}}{% else %}{{i.infile | ext}}{% endif %}"
	pTabix.args.tabix  = params.tabix.value
	pTabix.args.params = Box(h = True)
	pTabix.lang        = params.python.value
	pTabix.script      = "file:scripts/tabix/pTabix.py"
	return pTabix

@procfactory
def _pTabixIndex():
	"""
	@name:
		pTabixIndex
	@description:
		Generate tabix index file.
	@input:
		`infile:file`: the input file
			- Could be bgzipped.
	@output:
		`outfile:file`: The bgzipped file
		`outidx:file`:  The tabix index file
	@args:
		`tabix`: The path to `tabix`
		`params`: Other params for `tabix`
	"""
	pTabixIndex        = Proc(desc = 'Generate tabix index file')
	pTabixIndex.input  = "infile:file"
	pTabixIndex.output = [
		"outfile:file:{{i.infile | bn}}{% if args.gz %}.gz{% endif %}",
		"outidx:file:{{i.infile  | bn}}{% if args.gz %}.gz{% endif %}.tbi"
	]
	pTabixIndex.args.gz     = True
	pTabixIndex.args.tabix  = params.tabix.value
	pTabixIndex.args.params = Box()
	pTabixIndex.lang        = params.python.value
	pTabixIndex.script      = "file:scripts/tabix/pTabixIndex.py"
	return pTabixIndex




"""BigWig file operations using bwtool"""

from modkit import Modkit
from pyppl import Box, Proc
from . import params, delefactory, procfactory
Modkit().delegate(delefactory())

@procfactory
def _pBWExtract():
	"""
	@description:
		Extract features from bigwig file using `bwtool extract`.
		See: https://github.com/CRG-Barcelona/bwtool/wiki/extract
	@input:
		infile: The input bigwig file.
		bedfile: The bed/gff file of regions to extract features from.
	@output:
		outfile: The output features from those regions
	@args:
		bwtool (str): Path to bwtool.
	"""
	return Box(
		desc   = 'Extract features from bigwig file.',
		lang   = params.python.value,
		input  = 'infile:file, bedfile:file',
		output = 'outfile:file:{{i.infile | stem}}.bwtool-extracted.txt',
		args   = Box(bwtool = params.bwtool.value)
	)

@procfactory
def _pBWSummary():
	"""
	@description:
		Provide some summary stats for each region in a bed file using `bwtool summary`
		See: https://github.com/CRG-Barcelona/bwtool/wiki/summary
		Only bed file for regions is supported in this process.
	@input:
		infile: The input bigwig file.
		bedfile: The bed/gff file of regions to extract features from.
	@output:
		outfile: The output file.
	@args:
		bwtool (str): The path to bwtool
		params (Box): Other parameters for `bwtool summary`
	"""
	return Box(
		desc   = 'provide some summary stats for each region in a bed file',
		lang   = params.python.value,
		input  = 'infile:file, bedfile:file',
		output = 'outfile:file:{{i.infile | stem}}.bwtool-summarized.txt',
		args   = Box(
			bwtool = params.bwtool.value,
			params = Box(header = True, keep_bed = True)
		)
	)


"""EQTL analysis"""
from pyppl import Proc, Box
from . import params
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pMatrixeQTL():
	"""
	@name:
		pMatrixeQTL
	@description:
		Call eQTLs using Matrix eQTL
	@input:
		`snpfile:file`: The genotype file, rows are snps and columns are samples
		`expfile:file`: The expression file, rows are genes
		`covfile:file`: The covariant file, columns are covariants
	@output:
		`outfile:file`: The matrix eqtl output file
	@args:
		`model`: The model to use, either modelLINEAR(default) or modelANOVA
		`pval` : The pvalue cutoff (if `cisopts.dist` > 0, will be used as pval for trans-eQTL)
			- Set this to 0 and `cisopts.cispv` to do ciseqtl analysis only.
		`fdr`  : Calculate FDR or not (default: True)
		`cisopts`: Options for calling cis-, trans-eQTL
			- `snppos` : The snp position file (columns are: snp, chr, pos)
			- `genepos`: The gene position file (columns are: gene, chr, start, end)
			- `dist`   : The distance to define cis-eQTL. (default: 0 (don't do cis-, trans- calling)
			- `cispv`  : The pvalue cutoff for cis-eQTL (`pval` will not work). Default: `1e-3`
	@requires:
		[`Matrix-eQTL (R)`](http://www.bios.unc.edu/research/genomic_software/Matrix_eQTL/)
	"""
	pMatrixeQTL              = Proc(desc = 'Use matrix eQTL to call eQTLs')
	pMatrixeQTL.input        = 'snpfile:file, expfile:file, covfile:file'
	pMatrixeQTL.output       = 'outfile:file:{{i.snpfile | fn}}-{{i.expfile | fn}}.eqtl.txt, cisfile:file:{{i.snpfile | fn}}-{{i.expfile | fn}}.ciseqtl.txt'
	pMatrixeQTL.args.model   = 'modelLINEAR' # or modelANOVA
	pMatrixeQTL.args.pval    = 1e-5
	pMatrixeQTL.args.fdr     = True
	pMatrixeQTL.args.cisopts = Box(
		snppos  = '',
		genepos = params.refgene.value,
		dist    = 0,    # 0 don't do cis-, trans- calls
		cispv   = 1e-3
	)
	pMatrixeQTL.lang   = params.Rscript.value
	pMatrixeQTL.script = 'file:scripts/eqtl/pMatrixeQTL.r'
	return pMatrixeQTL




from os import path, makedirs
from glob import glob

{{runcmd}}

# split input bed file
beddir = path.join({{job.outdir | quote}}, 'chrom-regions')
if not path.isdir(beddir): makedirs(beddir)

bedfile = path.join({{job.outdir | quote}}, "{{i.bedfile | bn}}.sorted")
cmd = 'grep -v "^#" {{i.bedfile | quote}} | cut -f1,2,3,4,5,6  | sort -k1,1 -k2,2n > "%s"' % bedfile
runcmd(cmd)

chroms = []
with open(bedfile) as f:
	prevchr = ''
	handler = None
	for line in f:
		if not line.strip(): continue
		chrom = line.split('\t')[0]
		if chrom != prevchr:
			chroms.append(chrom)
			if handler: handler.close()
			handler = open(path.join(beddir, chrom + '.bed'), 'w')
			handler.write(line)
			prevchr = chrom
		else:
			handler.write(line)
	if handler:
		handler.close()

{% if args.pval %}
def getPval(num, dist):
	for i, d in enumerate(dist):
		if d < num: 
			return float(i)/float(len(dist))
	return 1.0

with open({{i.permfile | quote}}) as f:
	nulldist = [float(line.split('\t')[7].strip()) for line in f.read().splitlines() if line.strip()]
{% endif %}

# handle each chrom
outfileNoPval = "{{o.outfile}}.nopval"
for chrom in chroms:
	regfile = path.join(beddir, chrom + '.bed')
	bwfile  = glob(path.join({{args.consvdir | quote}}, chrom + '.*'))[0]
	cmd = '{{args.bwtool}} summary "%s" "%s" /dev/stdout -skip-median | cat >> "%s"' % (regfile, bwfile, outfileNoPval)
	runcmd (cmd)

outs = {}
with open(outfileNoPval) as f:
	for line in f:
		line  = line.strip()
		if not line: continue
		parts = line.split('\t')
		outs['\t'.join(parts[:3])] = 0 if parts[7] == 'NA' else float(parts[7])

with open({{i.bedfile | quote}}) as f, open({{out.outfile | quote}}, 'w') as fout:
	for line in f:
		line  = line.strip()
		if not line: continue
		if line.startswith('#'):
			{% if args.pval %}
			fout.write(line + '\tConservation\tConsvPval\n')
			{% else %}
			fout.write(line + '\tConservation\n')
			{% endif %}
			continue
		parts = line.split('\t')
		consv = outs['\t'.join(parts[:3])]
		{% if args.pval %}
		pval  = getPval(consv, nulldist)
		if pval >= {{args.pval}}: continue
		fout.write(line + '\t%.3f\t%.2E\n' % (consv, pval))
		{% else %}
		fout.write(line + '\t%.3f\n' % (consv))
		{% endif %}

